from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult

from .base import NotificationAction


@dataclass
class DeleteChannelAction(NotificationAction):
    """Action to delete a notification channel."""

    channel_id: UUID

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "delete_channel"

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.channel_id)


@dataclass
class DeleteChannelActionResult(BaseActionResult):
    """Result of deleting a notification channel."""

    deleted: bool

    @override
    def entity_id(self) -> Optional[str]:
        return None



import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import (
    Any,
    Optional,
)

from ai.backend.manager.models.vfolder import (
    QuotaScopeID,
    VFolderInvitationState,
    VFolderOperationStatus,
    VFolderOwnershipType,
    VFolderUsageMode,
)
from ai.backend.manager.models.vfolder import VFolderPermission as VFolderMountPermission


@dataclass
class VFolderBaseInfo:
    id: uuid.UUID
    quota_scope_id: Optional[QuotaScopeID]
    name: str
    host: str
    status: VFolderOperationStatus
    unmanaged_path: Optional[str]
    mount_permission: VFolderMountPermission
    usage_mode: VFolderUsageMode
    created_at: datetime
    cloneable: bool


@dataclass
class VFolderOwnershipInfo:
    creator_email: Optional[str]
    is_owner: bool
    ownership_type: VFolderOwnershipType
    user_uuid: Optional[uuid.UUID]
    group_uuid: Optional[uuid.UUID]


@dataclass
class VFolderUsageInfo:
    num_files: int
    used_bytes: int


@dataclass
class VFolderInvitationInfo:
    id: uuid.UUID
    vfolder_id: uuid.UUID
    vfolder_name: str
    invitee_user_email: str
    inviter_user_email: str
    mount_permission: VFolderMountPermission
    created_at: datetime
    modified_at: Optional[datetime]
    status: VFolderInvitationState

    def to_json(self) -> dict[str, Any]:
        return {
            "id": str(self.id),
            "vfolder_id": str(self.vfolder_id),
            "vfolder_name": self.vfolder_name,
            "invitee_user_email": self.invitee_user_email,
            "inviter_user_email": self.inviter_user_email,
            "mount_permission": self.mount_permission.value,
            "created_at": self.created_at.isoformat(),
            "modified_at": self.modified_at.isoformat() if self.modified_at is not None else None,
            "status": self.status.value,
        }


@dataclass
class FileInfo:
    name: str
    type: str
    size: int
    mode: str
    created: str
    modified: str

    def to_json(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "type": self.type,
            "size": self.size,
            "mode": self.mode,
            "created": self.created,
            "modified": self.modified,
        }



"""Add reservoir_registries table

Revision ID: 11ee5c074009
Revises: 4db74efaf82e
Create Date: 2025-08-26 03:31:43.771428

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "11ee5c074009"
down_revision = "4db74efaf82e"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "reservoir_registries",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("endpoint", sa.String(), nullable=False),
        sa.Column("access_key", sa.String(), nullable=False),
        sa.Column("secret_key", sa.String(), nullable=False),
        sa.Column("api_version", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_reservoir_registries")),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("reservoir_registries")
    # ### end Alembic commands ###



"""Main scheduling data type."""

from dataclasses import dataclass
from functools import cached_property
from typing import Optional

from ai.backend.common.types import ResourceSlot

from .agent import AgentMeta
from .base import SchedulingSpec
from .scaling_group import ScalingGroupMeta
from .session import PendingSessions
from .snapshot import SnapshotData


@dataclass
class SchedulingData:
    """Complete scheduling data structure."""

    scaling_group: ScalingGroupMeta
    pending_sessions: PendingSessions
    agents: list[AgentMeta]
    snapshot_data: Optional[SnapshotData]
    spec: SchedulingSpec

    @cached_property
    def total_capacity(self) -> ResourceSlot:
        """Calculate total available capacity from all agents."""
        return sum((agent.available_slots for agent in self.agents), ResourceSlot())



"""Base action class for app_config service."""

from __future__ import annotations

from ai.backend.manager.actions.action import BaseAction


class AppConfigAction(BaseAction):
    """Base class for all app_config actions."""



from __future__ import annotations

from .actions import ProcessNotificationAction, ProcessNotificationActionResult
from .processors import NotificationProcessors
from .service import NotificationService

__all__ = (
    "NotificationProcessors",
    "NotificationService",
    "ProcessNotificationAction",
    "ProcessNotificationActionResult",
)



"""Add verification_result to artifact_revisions

Revision ID: 4cfe35b1e60a
Revises: b55309328455
Create Date: 2025-11-08 01:53:01.604948

"""

import sqlalchemy as sa
from alembic import op

revision = "4cfe35b1e60a"
down_revision = "b55309328455"

branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "artifact_revisions",
        sa.Column("verification_result", sa.JSON(none_as_null=True), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("artifact_revisions", "verification_result")
    # ### end Alembic commands ###



import logging
from collections.abc import Mapping
from contextlib import AsyncExitStack
from dataclasses import dataclass
from typing import Optional

from ai.backend.common.clients.valkey_client.valkey_schedule import ValkeyScheduleClient
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.events.event_types.kernel.anycast import (
    KernelCancelledAnycastEvent,
    KernelCreatingAnycastEvent,
    KernelHeartbeatEvent,
    KernelPreparingAnycastEvent,
    KernelPullingAnycastEvent,
    KernelStartedAnycastEvent,
    KernelTerminatedAnycastEvent,
)
from ai.backend.common.events.event_types.schedule.anycast import (
    DoSokovanProcessIfNeededEvent,
    DoSokovanProcessScheduleEvent,
)
from ai.backend.common.leader.tasks import EventTaskSpec
from ai.backend.common.types import AgentId
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.metrics.scheduler import SchedulerOperationMetricObserver
from ai.backend.manager.scheduler.types import ScheduleType
from ai.backend.manager.sokovan.scheduler.scheduler import Scheduler
from ai.backend.manager.sokovan.scheduling_controller import SchedulingController
from ai.backend.manager.types import DistributedLockFactory

from .handlers import (
    CheckCreatingProgressHandler,
    CheckPreconditionHandler,
    CheckPullingProgressHandler,
    CheckRunningSessionTerminationHandler,
    CheckTerminatingProgressHandler,
    RetryCreatingHandler,
    RetryPreparingHandler,
    SchedulerHandler,
    ScheduleSessionsHandler,
    StartSessionsHandler,
    SweepLostAgentKernelsHandler,
    SweepSessionsHandler,
    SweepStaleKernelsHandler,
    TerminateSessionsHandler,
)
from .kernel import KernelStateEngine
from .recorder import RecorderContext
from .types import KernelCreationInfo

log = BraceStyleAdapter(logging.getLogger(__name__))


@dataclass
class SchedulerTaskSpec:
    """Specification for a scheduler's periodic task."""

    schedule_type: ScheduleType
    short_interval: Optional[float] = None  # None means no short-cycle task
    long_interval: float = 60.0
    initial_delay: float = 30.0

    def create_if_needed_event(self) -> DoSokovanProcessIfNeededEvent:
        """Create event for checking if processing is needed."""
        return DoSokovanProcessIfNeededEvent(self.schedule_type.value)

    def create_process_event(self) -> DoSokovanProcessScheduleEvent:
        """Create event for forced processing."""
        return DoSokovanProcessScheduleEvent(self.schedule_type.value)

    @property
    def short_task_name(self) -> str:
        """Name for the short-cycle task."""
        return f"sokovan_process_if_needed_{self.schedule_type.value}"

    @property
    def long_task_name(self) -> str:
        """Name for the long-cycle task."""
        return f"sokovan_process_schedule_{self.schedule_type.value}"


class ScheduleCoordinator:
    """
    Coordinate scheduling operations based on scheduling needs.
    Handles the actual scheduling logic and state management.
    """

    _valkey_schedule: ValkeyScheduleClient
    _scheduler: Scheduler
    _scheduling_controller: SchedulingController
    _schedule_handlers: Mapping[ScheduleType, SchedulerHandler]
    _operation_metrics: SchedulerOperationMetricObserver
    _kernel_state_engine: KernelStateEngine
    _lock_factory: DistributedLockFactory
    _config_provider: ManagerConfigProvider
    _event_producer: EventProducer

    def __init__(
        self,
        valkey_schedule: ValkeyScheduleClient,
        scheduler: Scheduler,
        scheduling_controller: SchedulingController,
        event_producer: EventProducer,
        lock_factory: DistributedLockFactory,
        config_provider: ManagerConfigProvider,
    ) -> None:
        self._valkey_schedule = valkey_schedule
        self._scheduler = scheduler
        self._scheduling_controller = scheduling_controller
        self._event_producer = event_producer
        self._lock_factory = lock_factory
        self._config_provider = config_provider
        self._operation_metrics = SchedulerOperationMetricObserver.instance()

        # Initialize kernel state engine with the scheduler's repository
        self._kernel_state_engine = KernelStateEngine(scheduler._repository)

        # Initialize handlers using a dedicated method
        self._schedule_handlers = self._init_handlers()

    def _init_handlers(self) -> Mapping[ScheduleType, SchedulerHandler]:
        """Initialize and return the mapping of schedule types to their handlers."""
        return {
            ScheduleType.SCHEDULE: ScheduleSessionsHandler(
                self._scheduler,
                self._scheduling_controller,
                self._event_producer,
                self._scheduler._repository,
            ),
            ScheduleType.CHECK_PRECONDITION: CheckPreconditionHandler(
                self._scheduler, self._scheduling_controller, self._event_producer
            ),
            ScheduleType.START: StartSessionsHandler(self._scheduler, self._event_producer),
            ScheduleType.TERMINATE: TerminateSessionsHandler(
                self._scheduler,
                self._scheduling_controller,
                self._event_producer,
                self._scheduler._repository,
            ),
            ScheduleType.SWEEP: SweepSessionsHandler(self._scheduler, self._scheduler._repository),
            ScheduleType.SWEEP_LOST_AGENT_KERNELS: SweepLostAgentKernelsHandler(
                self._scheduler, self._scheduler._repository
            ),
            ScheduleType.SWEEP_STALE_KERNELS: SweepStaleKernelsHandler(
                self._valkey_schedule, self._scheduler, self._scheduler._repository
            ),
            ScheduleType.CHECK_PULLING_PROGRESS: CheckPullingProgressHandler(
                self._scheduler, self._scheduling_controller, self._event_producer
            ),
            ScheduleType.CHECK_CREATING_PROGRESS: CheckCreatingProgressHandler(
                self._scheduler, self._scheduling_controller, self._event_producer
            ),
            ScheduleType.CHECK_TERMINATING_PROGRESS: CheckTerminatingProgressHandler(
                self._scheduler,
                self._scheduling_controller,
                self._event_producer,
                self._scheduler._repository,
            ),
            ScheduleType.CHECK_RUNNING_SESSION_TERMINATION: CheckRunningSessionTerminationHandler(
                self._valkey_schedule,
                self._scheduler,
                self._scheduler._repository,
            ),
            ScheduleType.RETRY_PREPARING: RetryPreparingHandler(
                self._scheduler, self._scheduling_controller, self._event_producer
            ),
            ScheduleType.RETRY_CREATING: RetryCreatingHandler(
                self._scheduler, self._scheduling_controller, self._event_producer
            ),
        }

    async def process_schedule(
        self,
        schedule_type: ScheduleType,
    ) -> bool:
        """
        Force processing of a specific schedule type.
        This method processes the scheduling operation even if it was not requested for guaranteed execution.
        So it should be called in long term loops.

        :param schedule_type: Type of scheduling operation
        :return: True if operation was performed, False otherwise
        """
        try:
            log.debug("Processing schedule type: {}", schedule_type.value)

            # Get handler from map and execute
            handler = self._schedule_handlers.get(schedule_type)
            if not handler:
                log.warning("No handler for schedule type: {}", schedule_type.value)
                return False

            # Execute the handler with optional locking and transition recording
            with RecorderContext.scope(schedule_type.value) as recorder:
                async with AsyncExitStack() as stack:
                    stack.enter_context(self._operation_metrics.measure_operation(handler.name()))
                    if handler.lock_id is not None:
                        lock_lifetime = (
                            self._config_provider.config.manager.session_schedule_lock_lifetime
                        )
                        await stack.enter_async_context(
                            self._lock_factory(handler.lock_id, lock_lifetime)
                        )
                    result = await handler.execute()
                    self._operation_metrics.observe_success(
                        operation=handler.name(), count=result.success_count()
                    )
                    if result.needs_post_processing():
                        try:
                            await handler.post_process(result)
                        except Exception as e:
                            log.error("Error during post-processing: {}", e)

                # Log recorded steps for debugging (can be saved to DB later)
                all_steps = recorder.get_all_steps()
                if all_steps:
                    log.debug(
                        "Recorded {} sessions with execution steps for {}",
                        len(all_steps),
                        schedule_type.value,
                    )
            return True
        except Exception as e:
            log.exception(
                "Error processing schedule type {}: {}",
                schedule_type.value,
                e,
            )
            raise

    async def process_if_needed(self, schedule_type: ScheduleType) -> bool:
        """
        Process scheduling operation if needed (based on internal state).
        This method checks if the scheduling operation was requested and processes it if so.

        :param schedule_type: Type of scheduling operation
        :return: True if operation was performed, False otherwise
        """
        # Check internal state (uses Redis marks)
        if not await self._valkey_schedule.load_and_delete_schedule_mark(schedule_type.value):
            return False

        return await self.process_schedule(schedule_type)

    # Kernel event handling methods using the coordinator's kernel state engine

    async def handle_kernel_pulling(self, event: KernelPullingAnycastEvent) -> bool:
        """Handle kernel pulling event through the kernel state engine."""
        result = await self._kernel_state_engine.mark_kernel_pulling(event.kernel_id, event.reason)
        if result:
            # Request CHECK_PULLING_PROGRESS to monitor image pull progress
            await self._scheduling_controller.mark_scheduling_needed(
                ScheduleType.CHECK_PULLING_PROGRESS
            )
        return result

    async def handle_kernel_creating(self, event: KernelCreatingAnycastEvent) -> bool:
        """Handle kernel creating event through the kernel state engine."""
        return await self._kernel_state_engine.mark_kernel_creating(event.kernel_id, event.reason)

    async def handle_kernel_running(self, event: KernelStartedAnycastEvent) -> bool:
        """Handle kernel running event through the kernel state engine."""
        # Convert event data to dataclass (always present, may be empty)
        creation_info = KernelCreationInfo.from_dict(dict(event.creation_info))
        result = await self._kernel_state_engine.mark_kernel_running(
            event.kernel_id,
            event.reason,
            creation_info,
        )
        if result:
            # Request CHECK_CREATING_PROGRESS to check if session should transition to RUNNING
            await self._scheduling_controller.mark_scheduling_needed(
                ScheduleType.CHECK_CREATING_PROGRESS
            )
        return result

    async def handle_kernel_preparing(self, event: KernelPreparingAnycastEvent) -> bool:
        """Handle kernel preparing event through the kernel state engine."""
        result = await self._kernel_state_engine.mark_kernel_preparing(event.kernel_id)
        if result:
            # Request CHECK_PRECONDITION to check if images are ready
            await self._scheduling_controller.mark_scheduling_needed(
                ScheduleType.CHECK_PRECONDITION
            )
        return result

    async def handle_kernel_cancelled(self, event: KernelCancelledAnycastEvent) -> bool:
        """Handle kernel cancelled event through the kernel state engine."""
        return await self._kernel_state_engine.mark_kernel_cancelled(
            event.kernel_id, event.session_id, event.reason
        )

    async def handle_kernel_terminated(self, event: KernelTerminatedAnycastEvent) -> bool:
        """Handle kernel terminated event through the kernel state engine."""
        log.info("Handling termination of kernel {}", event.kernel_id)
        result = await self._kernel_state_engine.mark_kernel_terminated(
            event.kernel_id, event.reason, event.exit_code
        )
        if result:
            # Request CHECK_RUNNING_SESSION_TERMINATION to check if RUNNING session should become TERMINATING
            await self._scheduling_controller.mark_scheduling_needed(
                ScheduleType.CHECK_RUNNING_SESSION_TERMINATION
            )
            # Request CHECK_TERMINATING_PROGRESS to check if session should transition to TERMINATED
            await self._scheduling_controller.mark_scheduling_needed(
                ScheduleType.CHECK_TERMINATING_PROGRESS
            )
        return result

    async def handle_kernel_heartbeat(self, event: KernelHeartbeatEvent) -> bool:
        """Handle kernel heartbeat event through the kernel state engine."""
        return await self._kernel_state_engine.update_kernel_heartbeat(event.kernel_id)

    # Image-related kernel state update methods

    async def update_kernels_to_pulling_for_image(
        self,
        agent_id: AgentId,
        image: str,
        image_ref: Optional[str] = None,
    ) -> None:
        """Update kernel status from PREPARING to PULLING for the specified image on an agent."""
        await self._kernel_state_engine.update_kernels_to_pulling_for_image(
            agent_id, image, image_ref
        )

    async def update_kernels_to_prepared_for_image(
        self,
        agent_id: AgentId,
        image: str,
        image_ref: Optional[str] = None,
    ) -> None:
        """Update kernel status to PREPARED for the specified image on an agent."""
        result = await self._kernel_state_engine.update_kernels_to_prepared_for_image(
            agent_id, image, image_ref
        )
        if result > 0:
            log.info(
                "Updated {} kernels to PREPARED state for agent:{} image:{}",
                result,
                agent_id,
                image,
            )
            # Request scheduling to check if sessions can transition to RUNNING
            await self._scheduling_controller.mark_scheduling_needed(
                ScheduleType.CHECK_PULLING_PROGRESS,
            )

    async def cancel_kernels_for_failed_image(
        self,
        agent_id: AgentId,
        image: str,
        error_msg: str,
        image_ref: Optional[str] = None,
    ) -> None:
        """Cancel kernels for an image that failed to be available on an agent."""
        await self._kernel_state_engine.cancel_kernels_for_failed_image(
            agent_id, image, error_msg, image_ref
        )
        # No need to request scheduling for cancelled kernels

    @staticmethod
    def _create_task_specs() -> list[SchedulerTaskSpec]:
        """Create task specifications for all schedule types."""
        return [
            # Regular scheduling operations with both short and long cycle tasks
            SchedulerTaskSpec(
                ScheduleType.SCHEDULE,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            SchedulerTaskSpec(
                ScheduleType.CHECK_PRECONDITION,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            SchedulerTaskSpec(
                ScheduleType.START,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            SchedulerTaskSpec(
                ScheduleType.TERMINATE,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            # Sweep is a maintenance task - only needs long cycle task
            SchedulerTaskSpec(
                ScheduleType.SWEEP,
                short_interval=None,  # No short-cycle task for maintenance
                long_interval=60.0,
                initial_delay=30.0,
            ),
            # Sweep lost agent kernels - maintenance task to clean up kernels with lost agents
            SchedulerTaskSpec(
                ScheduleType.SWEEP_LOST_AGENT_KERNELS,
                short_interval=None,  # No short-cycle task for maintenance
                long_interval=60.0,
                initial_delay=30.0,
            ),
            # Sweep stale kernels - maintenance task to clean up kernels with stale presence
            SchedulerTaskSpec(
                ScheduleType.SWEEP_STALE_KERNELS,
                short_interval=None,  # No short-cycle task for maintenance
                long_interval=60.0,
                initial_delay=30.0,
            ),
            # Progress check operations with both short and long cycle tasks
            SchedulerTaskSpec(
                ScheduleType.CHECK_PULLING_PROGRESS,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            SchedulerTaskSpec(
                ScheduleType.CHECK_CREATING_PROGRESS,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            SchedulerTaskSpec(
                ScheduleType.CHECK_TERMINATING_PROGRESS,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            # Check RUNNING sessions where all kernels are TERMINATED
            SchedulerTaskSpec(
                ScheduleType.CHECK_RUNNING_SESSION_TERMINATION,
                short_interval=2.0,
                long_interval=60.0,
                initial_delay=30.0,
            ),
            # Retry operations - only long cycle tasks
            SchedulerTaskSpec(
                ScheduleType.RETRY_PREPARING,
                short_interval=None,  # No short-cycle task
                long_interval=10.0,  # 10 seconds for retry operations
                initial_delay=10.0,  # Wait a bit before first retry
            ),
            SchedulerTaskSpec(
                ScheduleType.RETRY_CREATING,
                short_interval=None,  # No short-cycle task
                long_interval=10.0,  # 10 seconds for retry operations
                initial_delay=10.0,  # Wait a bit before first retry
            ),
        ]

    def create_task_specs(self) -> list[EventTaskSpec]:
        """Create task specifications for leader-based scheduling."""
        timer_specs = self._create_task_specs()
        specs: list[EventTaskSpec] = []

        for spec in timer_specs:
            # Create short-cycle task spec if specified
            if spec.short_interval is not None:
                short_spec = EventTaskSpec(
                    name=spec.short_task_name,
                    event_factory=spec.create_if_needed_event,
                    interval=spec.short_interval,
                    initial_delay=0.0,  # Start immediately for short tasks
                )
                specs.append(short_spec)

            # Create long-cycle task spec (always present)
            long_spec = EventTaskSpec(
                name=spec.long_task_name,
                event_factory=spec.create_process_event,
                interval=spec.long_interval,
                initial_delay=spec.initial_delay,
            )
            specs.append(long_spec)

        return specs



from dataclasses import dataclass, field
from typing import Any, Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.user.types import UserData
from ai.backend.manager.models.hasher.types import PasswordInfo
from ai.backend.manager.models.user import UserRole, UserStatus
from ai.backend.manager.services.user.actions.base import UserAction
from ai.backend.manager.types import OptionalState, PartialModifier, TriState


@dataclass
class UserModifier(PartialModifier):
    username: OptionalState[str] = field(default_factory=OptionalState.nop)
    password: OptionalState[PasswordInfo] = field(default_factory=OptionalState.nop)
    need_password_change: OptionalState[bool] = field(default_factory=OptionalState.nop)
    full_name: OptionalState[str] = field(default_factory=OptionalState.nop)
    description: OptionalState[str] = field(default_factory=OptionalState.nop)
    is_active: OptionalState[bool] = field(default_factory=OptionalState.nop)
    status: OptionalState[UserStatus] = field(default_factory=OptionalState.nop)
    domain_name: OptionalState[str] = field(default_factory=OptionalState.nop)
    role: OptionalState[UserRole] = field(default_factory=OptionalState.nop)
    allowed_client_ip: TriState[list[str]] = field(default_factory=TriState.nop)
    totp_activated: OptionalState[bool] = field(default_factory=OptionalState.nop)
    resource_policy: OptionalState[str] = field(default_factory=OptionalState.nop)
    sudo_session_enabled: OptionalState[bool] = field(default_factory=OptionalState.nop)
    main_access_key: TriState[str] = field(default_factory=TriState.nop)
    container_uid: TriState[int] = field(default_factory=TriState.nop)
    container_main_gid: TriState[int] = field(default_factory=TriState.nop)
    container_gids: TriState[list[int]] = field(default_factory=TriState.nop)
    group_ids: OptionalState[list[str]] = field(default_factory=OptionalState.nop)

    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.username.update_dict(to_update, "username")
        # Can't remove password
        password = self.password.optional_value()
        if password is not None:
            to_update["password"] = password
        self.need_password_change.update_dict(to_update, "need_password_change")
        self.full_name.update_dict(to_update, "full_name")
        self.description.update_dict(to_update, "description")
        self.is_active.update_dict(to_update, "is_active")
        self.domain_name.update_dict(to_update, "domain_name")
        self.role.update_dict(to_update, "role")
        self.allowed_client_ip.update_dict(to_update, "allowed_client_ip")
        self.totp_activated.update_dict(to_update, "totp_activated")
        self.resource_policy.update_dict(to_update, "resource_policy")
        self.sudo_session_enabled.update_dict(to_update, "sudo_session_enabled")
        self.main_access_key.update_dict(to_update, "main_access_key")
        self.container_uid.update_dict(to_update, "container_uid")
        self.container_main_gid.update_dict(to_update, "container_main_gid")
        self.container_gids.update_dict(to_update, "container_gids")
        # Set status based on is_active if not explicitly set
        status = self.status.optional_value()
        if status is None:
            is_active = self.is_active.optional_value()
            to_update["status"] = UserStatus.ACTIVE if is_active else UserStatus.INACTIVE
        else:
            to_update["status"] = status
        return to_update

    @property
    def group_ids_value(self) -> Optional[list[str]]:
        return self.group_ids.optional_value()


@dataclass
class ModifyUserAction(UserAction):
    email: str
    modifier: UserModifier

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "modify"


@dataclass
class ModifyUserActionResult(BaseActionResult):
    data: UserData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.data.id)



from typing import Callable, Optional

import sqlalchemy as sa

type QueryConditionCallable = Callable[
    [Optional[sa.sql.expression.BinaryExpression]], sa.sql.expression.BinaryExpression
]
type QueryCondition = Callable[[sa.sql.Select], QueryConditionCallable]

type QueryOptionCallable = Callable[[sa.sql.Select], sa.sql.Select]
type QueryOption = Callable[..., Callable[[sa.sql.Select], sa.sql.Select]]


def load_related_field(field: sa.orm.Load) -> QueryOptionCallable:
    return lambda stmt: stmt.options(field)


def join_by_related_field(field: sa.orm.attributes.InstrumentedAttribute) -> QueryOptionCallable:
    return lambda stmt: stmt.join(field)



import uuid
from typing import Optional

from pydantic import Field

from ai.backend.common.api_handlers import BaseRequestModel
from ai.backend.common.data.storage.registries.types import ModelSortKey, ModelTarget
from ai.backend.manager.data.artifact.modifier import ArtifactModifier
from ai.backend.manager.data.artifact.types import (
    ArtifactFilterOptions,
    ArtifactOrderingOptions,
    ArtifactType,
    DelegateeTarget,
)
from ai.backend.manager.defs import ARTIFACT_MAX_SCAN_LIMIT
from ai.backend.manager.types import PaginationOptions, TriState


class DelegateScanArtifactsReq(BaseRequestModel):
    delegator_reservoir_id: Optional[uuid.UUID] = Field(
        default=None, description="ID of the reservoir registry to delegate the scan request to"
    )
    delegatee_target: Optional[DelegateeTarget] = Field(
        default=None,
        description="The unique identifier of the target reservoir to delegate the scan.",
    )
    artifact_type: Optional[ArtifactType] = None
    limit: int = Field(
        lt=ARTIFACT_MAX_SCAN_LIMIT,
        description=f"Maximum number of artifacts to scan (max: {ARTIFACT_MAX_SCAN_LIMIT})",
    )
    order: Optional[ModelSortKey] = None
    search: Optional[str] = None


class ScanArtifactsReq(BaseRequestModel):
    registry_id: Optional[uuid.UUID] = Field(
        default=None, description="The unique identifier of the artifact registry to scan."
    )
    artifact_type: Optional[ArtifactType] = None
    limit: int = Field(
        lt=ARTIFACT_MAX_SCAN_LIMIT,
        description=f"Maximum number of artifacts to scan (max: {ARTIFACT_MAX_SCAN_LIMIT})",
    )
    order: Optional[ModelSortKey] = None
    search: Optional[str] = None


class ScanArtifactsSyncReq(BaseRequestModel):
    registry_id: Optional[uuid.UUID] = Field(
        default=None, description="The unique identifier of the artifact registry to scan."
    )
    artifact_type: Optional[ArtifactType] = None
    limit: int = Field(
        lt=ARTIFACT_MAX_SCAN_LIMIT,
        description=f"Maximum number of artifacts to scan (max: {ARTIFACT_MAX_SCAN_LIMIT})",
    )
    order: Optional[ModelSortKey] = None
    search: Optional[str] = None


class SearchArtifactsReq(BaseRequestModel):
    pagination: PaginationOptions
    ordering: Optional[ArtifactOrderingOptions] = None
    filters: Optional[ArtifactFilterOptions] = None


class ScanArtifactModelsReq(BaseRequestModel):
    models: list[ModelTarget] = Field(description="List of models to scan from the registry.")
    registry_id: Optional[uuid.UUID] = Field(
        default=None, description="The unique identifier of the artifact registry to scan."
    )


class ScanArtifactModelPathParam(BaseRequestModel):
    model_id: str = Field(description="The model to scan from the registry.")


class ScanArtifactModelQueryParam(BaseRequestModel):
    revision: Optional[str] = Field(description="The model revision to scan from the registry.")
    registry_id: Optional[uuid.UUID] = Field(
        default=None, description="The unique identifier of the artifact registry to scan."
    )


class CleanupArtifactsReq(BaseRequestModel):
    artifact_revision_ids: list[uuid.UUID] = Field(
        description="List of artifact revision IDs to cleanup."
    )


class CancelImportArtifactReq(BaseRequestModel):
    artifact_revision_id: uuid.UUID = Field(
        description="The artifact revision ID to cancel import."
    )


class ApproveArtifactRevisionReq(BaseRequestModel):
    artifact_revision_id: uuid.UUID = Field(description="The artifact revision ID to approve.")


class GetArtifactRevisionReadmeReq(BaseRequestModel):
    artifact_revision_id: uuid.UUID = Field(description="The artifact revision ID to get readme.")


class RejectArtifactRevisionReq(BaseRequestModel):
    artifact_revision_id: uuid.UUID = Field(description="The artifact revision ID to reject.")


class DelegateImportArtifactsReq(BaseRequestModel):
    artifact_revision_ids: list[uuid.UUID] = Field(
        description="List of artifact revision IDs to delegate the import request."
    )
    delegator_reservoir_id: Optional[uuid.UUID] = Field(
        default=None, description="ID of the reservoir registry to delegate the import request to"
    )
    delegatee_target: Optional[DelegateeTarget] = Field(
        default=None,
    )
    artifact_type: Optional[ArtifactType]


class ImportArtifactsReq(BaseRequestModel):
    artifact_revision_ids: list[uuid.UUID] = Field(
        description="List of artifact revision IDs to import."
    )


class UpdateArtifactReqPathParam(BaseRequestModel):
    artifact_id: uuid.UUID = Field(description="The artifact ID to update.")


class UpdateArtifactReqBodyParam(BaseRequestModel):
    readonly: Optional[bool] = Field(
        default=None, description="Whether the artifact should be readonly."
    )
    description: Optional[str] = Field(default=None, description="Updated description")

    def to_modifier(self) -> ArtifactModifier:
        modifier = ArtifactModifier()
        if self.readonly is not None:
            modifier.readonly = TriState.update(self.readonly)
        if self.description is not None:
            modifier.description = TriState.update(self.description)
        return modifier


class GetDownloadProgressReqPathParam(BaseRequestModel):
    artifact_revision_id: uuid.UUID = Field(
        description="The artifact revision ID to get download progress for"
    )



"""
Dispersed agent selector implementation for sokovan scheduler.

This selector prefers agents with more available resources to spread
workloads across the cluster.
"""

import sys
from decimal import Decimal
from typing import Sequence, Union

from .selector import (
    AbstractAgentSelector,
    AgentSelectionConfig,
    AgentSelectionCriteria,
    AgentStateTracker,
    ResourceRequirements,
)
from .utils import count_unutilized_capabilities, order_slots_by_priority


class DispersedAgentSelector(AbstractAgentSelector):
    """
    Dispersed agent selector that spreads workloads across agents.

    This selector prefers agents with:
    1. Fewer unutilized capabilities
    2. More available resources (to spread workloads)
    """

    def __init__(self, agent_selection_resource_priority: list[str]) -> None:
        """
        Initialize the dispersed selector.

        Args:
            agent_selection_resource_priority: Resource types in priority order
        """
        self.agent_selection_resource_priority = agent_selection_resource_priority

    def name(self) -> str:
        """
        Return the selector name for predicates.
        """
        return "DispersedAgentSelector"

    def success_message(self) -> str:
        """
        Return a message describing successful agent selection.
        """
        return "Agent selected using dispersed strategy for balanced workload distribution"

    def select_tracker_by_strategy(
        self,
        trackers: Sequence[AgentStateTracker],
        resource_req: ResourceRequirements,
        criteria: AgentSelectionCriteria,
        config: AgentSelectionConfig,
    ) -> AgentStateTracker:
        """
        Select an agent tracker to disperse workloads.

        Assumes trackers are already filtered for compatibility.
        """
        # Sort requested slots by priority
        resource_priorities = order_slots_by_priority(
            resource_req.requested_slots, self.agent_selection_resource_priority
        )

        # Choose the tracker with maximum available resources (to disperse workloads)
        def tracker_sort_key(tracker: AgentStateTracker) -> list[Union[int, Decimal]]:
            occupied_slots = tracker.get_current_occupied_slots()
            return [
                # First, prefer agents with fewer unutilized capabilities
                -count_unutilized_capabilities(
                    tracker.original_agent, resource_req.requested_slots
                ),
                # Then, prefer agents with more available resources (using current state)
                *[
                    (tracker.original_agent.available_slots - occupied_slots).get(key, -sys.maxsize)
                    for key in resource_priorities
                ],
            ]

        chosen_tracker = max(trackers, key=tracker_sort_key)

        return chosen_tracker



"""Add object_storages table

Revision ID: e38044633465
Revises: 42feff246198
Create Date: 2025-08-13 06:15:55.574744

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "e38044633465"
down_revision = "42feff246198"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "object_storages",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("host", sa.String(), nullable=False),
        sa.Column("access_key", sa.String(), nullable=False),
        sa.Column("secret_key", sa.String(), nullable=False),
        sa.Column("endpoint", sa.String(), nullable=False),
        sa.Column("region", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_object_storages")),
    )
    op.create_index(op.f("ix_object_storages_host"), "object_storages", ["host"], unique=False)
    op.create_index(op.f("ix_object_storages_name"), "object_storages", ["name"], unique=True)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_object_storages_name"), table_name="object_storages")
    op.drop_index(op.f("ix_object_storages_host"), table_name="object_storages")
    op.drop_table("object_storages")
    # ### end Alembic commands ###



from . import domain as _domain
from . import group as _group
from . import image as _image
from . import kernel as _kernel
from . import keypair as _keypair
from . import resource_preset as _resource_preset
from . import scaling_group as _scaling_group
from . import session as _session
from . import user as _user

__all__ = (
    *_domain.__all__,
    *_group.__all__,
    *_image.__all__,
    *_kernel.__all__,
    *_keypair.__all__,
    *_resource_preset.__all__,
    *_scaling_group.__all__,
    *_session.__all__,
    *_user.__all__,
)

from .domain import *  # noqa
from .group import *  # noqa
from .image import *  # noqa
from .kernel import *  # noqa
from .keypair import *  # noqa
from .resource_preset import *  # noqa
from .scaling_group import *  # noqa
from .session import *  # noqa
from .user import *  # noqa



"""
REST-style session management APIs.
"""

from __future__ import annotations

import asyncio
import functools
import json
import logging
from collections.abc import Iterable, Mapping
from datetime import datetime, timedelta
from decimal import Decimal
from http import HTTPStatus
from typing import TYPE_CHECKING, Annotated, Any, Optional, cast
from uuid import UUID

import aiohttp_cors
import aiotools
import attrs
import sqlalchemy as sa
import sqlalchemy.exc
import trafaret as t
from aiohttp import web
from aiotools import cancel_and_wait
from dateutil.tz import tzutc
from pydantic import AliasChoices, Field
from sqlalchemy.sql.expression import null, true

from ai.backend.common.data.session.types import CustomizedImageVisibilityScope
from ai.backend.common.json import read_json
from ai.backend.manager.services.agent.actions.sync_agent_registry import SyncAgentRegistryAction
from ai.backend.manager.services.session.actions.check_and_transit_status import (
    CheckAndTransitStatusAction,
)
from ai.backend.manager.services.session.actions.commit_session import CommitSessionAction
from ai.backend.manager.services.session.actions.complete import CompleteAction
from ai.backend.manager.services.session.actions.convert_session_to_image import (
    ConvertSessionToImageAction,
)
from ai.backend.manager.services.session.actions.create_cluster import CreateClusterAction
from ai.backend.manager.services.session.actions.create_from_params import (
    CreateFromParamsAction,
    CreateFromParamsActionParams,
)
from ai.backend.manager.services.session.actions.create_from_template import (
    CreateFromTemplateAction,
    CreateFromTemplateActionParams,
)
from ai.backend.manager.services.session.actions.destroy_session import DestroySessionAction
from ai.backend.manager.services.session.actions.download_file import DownloadFileAction
from ai.backend.manager.services.session.actions.download_files import DownloadFilesAction
from ai.backend.manager.services.session.actions.execute_session import (
    ExecuteSessionAction,
    ExecuteSessionActionParams,
)
from ai.backend.manager.services.session.actions.get_abusing_report import GetAbusingReportAction
from ai.backend.manager.services.session.actions.get_commit_status import GetCommitStatusAction
from ai.backend.manager.services.session.actions.get_container_logs import GetContainerLogsAction
from ai.backend.manager.services.session.actions.get_dependency_graph import (
    GetDependencyGraphAction,
)
from ai.backend.manager.services.session.actions.get_direct_access_info import (
    GetDirectAccessInfoAction,
)
from ai.backend.manager.services.session.actions.get_session_info import GetSessionInfoAction
from ai.backend.manager.services.session.actions.get_status_history import GetStatusHistoryAction
from ai.backend.manager.services.session.actions.interrupt_session import InterruptSessionAction
from ai.backend.manager.services.session.actions.list_files import ListFilesAction
from ai.backend.manager.services.session.actions.match_sessions import MatchSessionsAction
from ai.backend.manager.services.session.actions.rename_session import RenameSessionAction
from ai.backend.manager.services.session.actions.restart_session import RestartSessionAction
from ai.backend.manager.services.session.actions.shutdown_service import ShutdownServiceAction
from ai.backend.manager.services.session.actions.start_service import StartServiceAction
from ai.backend.manager.services.session.actions.upload_files import UploadFilesAction
from ai.backend.manager.services.vfolder.actions.base import GetTaskLogsAction

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
    from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.common import validators as tx
from ai.backend.common.defs.session import (
    SESSION_PRIORITY_DEFAULT,
    SESSION_PRIORITY_MAX,
    SESSION_PRIORITY_MIN,
)
from ai.backend.common.events.event_types.agent.anycast import (
    AgentTerminatedEvent,
)
from ai.backend.common.exception import BackendAIError
from ai.backend.common.plugin.monitor import GAUGE
from ai.backend.common.types import (
    AccessKey,
    AgentId,
    ClusterMode,
    KernelId,
    MountPermission,
    MountTypes,
    SessionId,
    SessionTypes,
)
from ai.backend.logging import BraceStyleAdapter

from ..defs import DEFAULT_IMAGE_ARCH, DEFAULT_ROLE
from ..errors.api import InvalidAPIParameters
from ..errors.auth import InsufficientPrivilege
from ..errors.kernel import InvalidSessionData, SessionNotFound
from ..errors.resource import NoCurrentTaskContext
from ..models import (
    AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES,
    SessionDependencyRow,
    SessionRow,
    UserRole,
    kernels,
    keypairs,
)
from ..utils import query_userinfo as _query_userinfo
from .auth import auth_required
from .manager import ALL_ALLOWED, READ_ALLOWED, server_status_required
from .types import CORSOptions, WebMiddleware
from .utils import (
    LegacyBaseRequestModel,
    LegacyBaseResponseModel,
    Undefined,
    catch_unexpected,
    check_api_params,
    deprecated_stub,
    get_access_key_scopes,
    pydantic_params_api_handler,
    undefined,
)

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

_json_loads = functools.partial(json.loads, parse_float=Decimal)


class UndefChecker(t.Trafaret):
    def check_and_return(self, value: Any) -> object:
        if value == undefined:
            return value
        else:
            self._failure("Invalid Undef format", value=value)
            return None


resource_opts_iv = t.Dict({
    t.Key("shmem", default=None): t.Null | tx.BinarySize,
    t.Key("allow_fractional_resource_fragmentation", default=None): t.Null | t.ToBool,
}).allow_extra("*")

creation_config_v1 = t.Dict({
    t.Key("mounts", default=None): t.Null | t.List(t.String),
    t.Key("environ", default=None): t.Null | t.Mapping(t.String, t.String),
    t.Key("clusterSize", default=None): t.Null | t.Int[1:],
})
creation_config_v2 = t.Dict({
    t.Key("mounts", default=None): t.Null | t.List(t.String),
    t.Key("environ", default=None): t.Null | t.Mapping(t.String, t.String),
    t.Key("clusterSize", default=None): t.Null | t.Int[1:],
    t.Key("instanceMemory", default=None): t.Null | tx.BinarySize,
    t.Key("instanceCores", default=None): t.Null | t.Int,
    t.Key("instanceGPUs", default=None): t.Null | t.Float,
    t.Key("instanceTPUs", default=None): t.Null | t.Int,
})
creation_config_v3 = t.Dict({
    t.Key("mounts", default=None): t.Null | t.List(t.String),
    t.Key("environ", default=None): t.Null | t.Mapping(t.String, t.String),
    tx.AliasedKey(["cluster_size", "clusterSize"], default=None): t.Null | t.Int[1:],
    tx.AliasedKey(["scaling_group", "scalingGroup"], default=None): t.Null | t.String,
    t.Key("resources", default=None): t.Null | t.Mapping(t.String, t.Any),
    tx.AliasedKey(["resource_opts", "resourceOpts"], default=None): t.Null | resource_opts_iv,
})
creation_config_v3_template = t.Dict({
    t.Key("mounts", default=undefined): UndefChecker | t.Null | t.List(t.String),
    t.Key("environ", default=undefined): UndefChecker | t.Null | t.Mapping(t.String, t.String),
    tx.AliasedKey(["cluster_size", "clusterSize"], default=undefined): (
        UndefChecker | t.Null | t.Int[1:]
    ),
    tx.AliasedKey(["scaling_group", "scalingGroup"], default=undefined): (
        UndefChecker | t.Null | t.String
    ),
    t.Key("resources", default=undefined): UndefChecker | t.Null | t.Mapping(t.String, t.Any),
    tx.AliasedKey(["resource_opts", "resourceOpts"], default=undefined): (
        UndefChecker | resource_opts_iv
    ),
})
creation_config_v4 = t.Dict({
    t.Key("mounts", default=None): t.Null | t.List(t.String),
    tx.AliasedKey(["mount_map", "mountMap"], default=None): t.Null | t.Mapping(t.String, t.String),
    t.Key("environ", default=None): t.Null | t.Mapping(t.String, t.String),
    tx.AliasedKey(["cluster_size", "clusterSize"], default=None): t.Null | t.Int[1:],
    tx.AliasedKey(["scaling_group", "scalingGroup"], default=None): t.Null | t.String,
    t.Key("resources", default=None): t.Null | t.Mapping(t.String, t.Any),
    tx.AliasedKey(["resource_opts", "resourceOpts"], default=None): t.Null | resource_opts_iv,
    tx.AliasedKey(["preopen_ports", "preopenPorts"], default=None): t.Null
    | t.List(t.Int[1024:65535]),
})
creation_config_v4_template = t.Dict({
    t.Key("mounts", default=undefined): UndefChecker | t.Null | t.List(t.String),
    tx.AliasedKey(["mount_map", "mountMap"], default=undefined): (
        UndefChecker | t.Null | t.Mapping(t.String, t.String)
    ),
    t.Key("environ", default=undefined): UndefChecker | t.Null | t.Mapping(t.String, t.String),
    tx.AliasedKey(["cluster_size", "clusterSize"], default=undefined): (
        UndefChecker | t.Null | t.Int[1:]
    ),
    tx.AliasedKey(["scaling_group", "scalingGroup"], default=undefined): (
        UndefChecker | t.Null | t.String
    ),
    t.Key("resources", default=undefined): UndefChecker | t.Null | t.Mapping(t.String, t.Any),
    tx.AliasedKey(["resource_opts", "resourceOpts"], default=undefined): (
        UndefChecker | t.Null | resource_opts_iv
    ),
})
creation_config_v5 = t.Dict({
    t.Key("mounts", default=None): t.Null | t.List(t.String),
    tx.AliasedKey(["mount_map", "mountMap"], default=None): t.Null | t.Mapping(t.String, t.String),
    tx.AliasedKey(["mount_options", "mountOptions"], default=None): t.Null
    | t.Mapping(
        t.String,
        t.Dict({
            t.Key("type", default=MountTypes.BIND): tx.Enum(MountTypes),
            tx.AliasedKey(["permission", "perm"], default=None): t.Null | tx.Enum(MountPermission),
        }).ignore_extra("*"),
    ),
    t.Key("environ", default=None): t.Null | t.Mapping(t.String, t.String),
    # cluster_size is moved to the root-level parameters
    tx.AliasedKey(["scaling_group", "scalingGroup"], default=None): t.Null | t.String,
    t.Key("resources", default=None): t.Null | t.Mapping(t.String, t.Any),
    tx.AliasedKey(["resource_opts", "resourceOpts"], default=None): t.Null | resource_opts_iv,
    tx.AliasedKey(["preopen_ports", "preopenPorts"], default=None): t.Null
    | t.List(t.Int[1024:65535]),
    tx.AliasedKey(["agent_list", "agentList"], default=None): t.Null | t.List(t.String),
})
creation_config_v5_template = t.Dict({
    t.Key("mounts", default=undefined): UndefChecker | t.Null | t.List(t.String),
    tx.AliasedKey(["mount_map", "mountMap"], default=undefined): (
        UndefChecker | t.Null | t.Mapping(t.String, t.String)
    ),
    t.Key("environ", default=undefined): UndefChecker | t.Null | t.Mapping(t.String, t.String),
    # cluster_size is moved to the root-level parameters
    tx.AliasedKey(["scaling_group", "scalingGroup"], default=undefined): (
        UndefChecker | t.Null | t.String
    ),
    t.Key("resources", default=undefined): UndefChecker | t.Null | t.Mapping(t.String, t.Any),
    tx.AliasedKey(["resource_opts", "resourceOpts"], default=undefined): (
        UndefChecker | t.Null | resource_opts_iv
    ),
})
creation_config_v6 = t.Dict({
    t.Key("mounts", default=None): t.Null | t.List(t.String),
    tx.AliasedKey(["mount_map", "mountMap"], default=None): t.Null | t.Mapping(t.String, t.String),
    tx.AliasedKey(["mount_options", "mountOptions"], default=None): t.Null
    | t.Mapping(
        t.String,
        t.Dict({
            t.Key("type", default=MountTypes.BIND): tx.Enum(MountTypes),
            tx.AliasedKey(["permission", "perm"], default=None): t.Null | tx.Enum(MountPermission),
        }).ignore_extra("*"),
    ),
    t.Key("environ", default=None): t.Null | t.Mapping(t.String, t.String),
    # cluster_size is moved to the root-level parameters
    tx.AliasedKey(["scaling_group", "scalingGroup"], default=None): t.Null | t.String,
    t.Key("resources", default=None): t.Null | t.Mapping(t.String, t.Any),
    tx.AliasedKey(["resource_opts", "resourceOpts"], default=None): t.Null | resource_opts_iv,
    tx.AliasedKey(["preopen_ports", "preopenPorts"], default=None): t.Null
    | t.List(t.Int[1024:65535]),
    tx.AliasedKey(["agent_list", "agentList"], default=None): t.Null | t.List(t.String),
    tx.AliasedKey(["attach_network", "attachNetwork"], default=None): t.Null | tx.UUID,
})
creation_config_v6_template = t.Dict({
    t.Key("mounts", default=undefined): UndefChecker | t.Null | t.List(t.String),
    tx.AliasedKey(["mount_map", "mountMap"], default=undefined): (
        UndefChecker | t.Null | t.Mapping(t.String, t.String)
    ),
    t.Key("environ", default=undefined): UndefChecker | t.Null | t.Mapping(t.String, t.String),
    # cluster_size is moved to the root-level parameters
    tx.AliasedKey(["scaling_group", "scalingGroup"], default=undefined): (
        UndefChecker | t.Null | t.String
    ),
    t.Key("resources", default=undefined): UndefChecker | t.Null | t.Mapping(t.String, t.Any),
    tx.AliasedKey(["resource_opts", "resourceOpts"], default=undefined): (
        UndefChecker | t.Null | resource_opts_iv
    ),
    tx.AliasedKey(["attach_network", "attachNetwork"], default=undefined): (
        UndefChecker | t.Null | tx.UUID
    ),
})

creation_config_v7 = t.Dict({
    t.Key("mounts", default=None): t.Null | t.List(t.String),  # deprecated
    tx.AliasedKey(["mount_map", "mountMap"], default=None): t.Null
    | t.Mapping(t.String, t.String),  # deprecated
    t.Key("mount_ids", default=None): t.Null | t.List(tx.UUID),
    tx.AliasedKey(["mount_id_map", "mountIdMap"], default=None): t.Null
    | t.Mapping(tx.UUID, t.String),
    tx.AliasedKey(["mount_options", "mountOptions"], default=None): t.Null
    | t.Mapping(
        t.String,
        t.Dict({
            t.Key("type", default=MountTypes.BIND): tx.Enum(MountTypes),
            tx.AliasedKey(["permission", "perm"], default=None): t.Null | tx.Enum(MountPermission),
        }).ignore_extra("*"),
    ),
    t.Key("environ", default=None): t.Null | t.Mapping(t.String, t.String),
    # cluster_size is moved to the root-level parameters
    tx.AliasedKey(["scaling_group", "scalingGroup"], default=None): t.Null | t.String,
    t.Key("resources", default=None): t.Null | t.Mapping(t.String, t.Any),
    tx.AliasedKey(["resource_opts", "resourceOpts"], default=None): t.Null | resource_opts_iv,
    tx.AliasedKey(["preopen_ports", "preopenPorts"], default=None): t.Null
    | t.List(t.Int[1024:65535]),
    tx.AliasedKey(["agent_list", "agentList"], default=None): t.Null | t.List(t.String),
    tx.AliasedKey(["attach_network", "attachNetwork"], default=None): t.Null | tx.UUID,
})

overwritten_param_check = t.Dict({
    t.Key("template_id"): tx.UUID,
    t.Key("session_name"): tx.SessionName,
    t.Key("image", default=None): t.Null | t.String,
    tx.AliasedKey(["session_type", "sess_type"]): tx.Enum(SessionTypes),
    t.Key("group", default=None): t.Null | t.String,
    t.Key("domain", default=None): t.Null | t.String,
    t.Key("config", default=None): t.Null | t.Mapping(t.String, t.Any),
    t.Key("tag", default=None): t.Null | t.String,
    t.Key("enqueue_only", default=False): t.ToBool,
    t.Key("max_wait_seconds", default=0): t.Int[0:],
    t.Key("reuse", default=True): t.ToBool,
    t.Key("startup_command", default=None): t.Null | t.String,
    t.Key("bootstrap_script", default=None): t.Null | t.String,
    t.Key("owner_access_key", default=None): t.Null | t.String,
    tx.AliasedKey(["scaling_group", "scalingGroup"], default=None): t.Null | t.String,
    tx.AliasedKey(["cluster_size", "clusterSize"], default=None): t.Null | t.Int[1:],
    tx.AliasedKey(["cluster_mode", "clusterMode"], default="SINGLE_NODE"): tx.Enum(ClusterMode),
    tx.AliasedKey(["starts_at", "startsAt"], default=None): t.Null | t.String,
    tx.AliasedKey(["batch_timeout", "batchTimeout"], default=None): t.Null | tx.TimeDuration,
}).allow_extra("*")


def sub(d, old, new):
    for k, v in d.items():
        if isinstance(v, Mapping) or isinstance(v, dict):
            d[k] = sub(v, old, new)
        elif d[k] == old:
            d[k] = new
    return d


def drop_undefined(d):
    newd = {}
    for k, v in d.items():
        if isinstance(v, Mapping) or isinstance(v, dict):
            newval = drop_undefined(v)
            if len(newval.keys()) > 0:  # exclude empty dict always
                newd[k] = newval
        elif not isinstance(v, Undefined):
            newd[k] = v
    return newd


async def query_userinfo(
    request: web.Request,
    params: Any,
    conn: SAConnection,
) -> tuple[UUID, UUID, dict]:
    try:
        return await _query_userinfo(
            conn,
            request["user"]["uuid"],
            request["keypair"]["access_key"],
            request["user"]["role"],
            request["user"]["domain_name"],
            request["keypair"]["resource_policy"],
            params["domain"] or request["user"]["domain_name"],
            params["group"],
            query_on_behalf_of=(
                None if params["owner_access_key"] is undefined else params["owner_access_key"]
            ),
        )
    except ValueError as e:
        raise InvalidAPIParameters(str(e))


@server_status_required(ALL_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(["template_id", "templateId"]): t.Null | tx.UUID,
        tx.AliasedKey(["name", "session_name", "clientSessionToken"], default=undefined)
        >> "session_name": UndefChecker | tx.SessionName,
        t.Key("priority", default=SESSION_PRIORITY_DEFAULT): t.ToInt(
            gte=SESSION_PRIORITY_MIN, lte=SESSION_PRIORITY_MAX
        ),
        tx.AliasedKey(["image", "lang"], default=undefined): UndefChecker | t.Null | t.String,
        tx.AliasedKey(["arch", "architecture"], default=undefined) >> "architecture": t.String
        | UndefChecker,
        tx.AliasedKey(["type", "sessionType"], default=undefined) >> "session_type": tx.Enum(
            SessionTypes
        )
        | UndefChecker,
        tx.AliasedKey(["group", "groupName", "group_name"], default=undefined): (
            UndefChecker | t.Null | t.String
        ),
        tx.AliasedKey(["domain", "domainName", "domain_name"], default=undefined): (
            UndefChecker | t.Null | t.String
        ),
        tx.AliasedKey(["cluster_size", "clusterSize"], default=1): t.ToInt[1:],  # new in APIv6
        tx.AliasedKey(["cluster_mode", "clusterMode"], default="SINGLE_NODE"): tx.Enum(
            ClusterMode
        ),  # new in APIv6
        t.Key("config", default=dict): t.Mapping(t.String, t.Any),
        t.Key("tag", default=undefined): UndefChecker | t.Null | t.String,
        t.Key("enqueueOnly", default=False) >> "enqueue_only": t.ToBool,
        t.Key("maxWaitSeconds", default=0) >> "max_wait_seconds": t.Int[0:],
        tx.AliasedKey(["starts_at", "startsAt"], default=None): t.Null | t.String,
        tx.AliasedKey(["batch_timeout", "batchTimeout"], default=None): t.Null | tx.TimeDuration,
        t.Key("reuseIfExists", default=True) >> "reuse": t.ToBool,
        t.Key("startupCommand", default=None) >> "startup_command": UndefChecker
        | t.Null
        | t.String,
        tx.AliasedKey(["bootstrap_script", "bootstrapScript"], default=undefined): (
            UndefChecker | t.Null | t.String
        ),
        t.Key("dependencies", default=None): (
            UndefChecker | t.Null | t.List(tx.UUID) | t.List(t.String)
        ),
        tx.AliasedKey(["callback_url", "callbackUrl", "callbackURL"], default=None): (
            UndefChecker | t.Null | tx.URL
        ),
        t.Key("owner_access_key", default=undefined): UndefChecker | t.Null | t.String,
    }),
    loads=_json_loads,
)
async def create_from_template(request: web.Request, params: dict[str, Any]) -> web.Response:
    # TODO: we need to refactor session_template model to load the template configs
    #       by one batch. Currently, we need to set every template configs one by one.
    root_ctx: RootContext = request.app["_root.context"]

    if params["image"] is None and params["template_id"] is None:
        raise InvalidAPIParameters("Both image and template_id can't be None!")

    api_version = request["api_version"]
    try:
        if 8 <= api_version[0]:
            params["config"] = creation_config_v6_template.check(params["config"])
        elif 6 <= api_version[0]:
            params["config"] = creation_config_v5_template.check(params["config"])
        elif 5 <= api_version[0]:
            params["config"] = creation_config_v4_template.check(params["config"])
        elif (4, "20190315") <= api_version:
            params["config"] = creation_config_v3_template.check(params["config"])
    except t.DataError as e:
        log.debug("Validation error: {0}", e.as_dict())
        raise InvalidAPIParameters("Input validation error", extra_data=e.as_dict())

    scopes_param = {
        "owner_access_key": (
            None if params["owner_access_key"] is undefined else params["owner_access_key"]
        ),
    }

    requester_access_key, owner_access_key = await get_access_key_scopes(request, scopes_param)
    log.info(
        "GET_OR_CREATE (ak:{0}/{1}, img:{2}, s:{3})",
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else "*",
        params["image"],
        params["session_name"],
    )

    if params["domain"] is None:
        domain_name = request["user"]["domain_name"]
    else:
        domain_name = params["domain"]

    result = await root_ctx.processors.session.create_from_template.wait_for_complete(
        CreateFromTemplateAction(
            params=CreateFromTemplateActionParams(
                template_id=params["template_id"],
                session_name=params["session_name"],
                image=params["image"],
                architecture=params["architecture"],
                session_type=params["session_type"],
                group_name=params["group"],
                domain_name=domain_name,
                cluster_size=params["cluster_size"],
                cluster_mode=params["cluster_mode"],
                config=params["config"],
                tag=params["tag"],
                enqueue_only=params["enqueue_only"],
                max_wait_seconds=params["max_wait_seconds"],
                reuse_if_exists=params["reuse"],
                startup_command=params["startup_command"],
                bootstrap_script=params["bootstrap_script"],
                dependencies=params["dependencies"],
                callback_url=params["callback_url"],
                priority=params["priority"],
                starts_at=params["starts_at"],
                batch_timeout=params["batch_timeout"],
                owner_access_key=owner_access_key,
            ),
            user_id=request["user"]["uuid"],
            user_role=request["user"]["role"],
            requester_access_key=requester_access_key,
            sudo_session_enabled=request["user"]["sudo_session_enabled"],
            keypair_resource_policy=request["keypair"]["resource_policy"],
        )
    )

    return web.json_response(result.result, status=HTTPStatus.CREATED)


@server_status_required(ALL_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(["name", "session_name", "clientSessionToken"])
        >> "session_name": tx.SessionName,
        t.Key("priority", default=SESSION_PRIORITY_DEFAULT): t.ToInt(
            gte=SESSION_PRIORITY_MIN, lte=SESSION_PRIORITY_MAX
        ),
        tx.AliasedKey(["image", "lang"]): t.String,
        tx.AliasedKey(["arch", "architecture"], default=DEFAULT_IMAGE_ARCH)
        >> "architecture": t.String,
        tx.AliasedKey(["type", "sessionType"], default="interactive") >> "session_type": tx.Enum(
            SessionTypes
        ),
        tx.AliasedKey(["group", "groupName", "group_name"], default="default"): t.String,
        tx.AliasedKey(["domain", "domainName", "domain_name"], default="default"): t.String,
        tx.AliasedKey(["cluster_size", "clusterSize"], default=1): t.ToInt[1:],  # new in APIv6
        tx.AliasedKey(["cluster_mode", "clusterMode"], default="SINGLE_NODE"): tx.Enum(
            ClusterMode
        ),  # new in APIv6
        t.Key("config", default=dict): t.Mapping(t.String, t.Any),
        t.Key("tag", default=None): t.Null | t.String,
        t.Key("enqueueOnly", default=False) >> "enqueue_only": t.ToBool,
        t.Key("maxWaitSeconds", default=0) >> "max_wait_seconds": t.ToInt[0:],
        tx.AliasedKey(["starts_at", "startsAt"], default=None): t.Null | t.String,
        tx.AliasedKey(["batch_timeout", "batchTimeout"], default=None): t.Null | tx.TimeDuration,
        t.Key("reuseIfExists", default=True) >> "reuse": t.ToBool,
        t.Key("startupCommand", default=None) >> "startup_command": t.Null | t.String,
        tx.AliasedKey(["bootstrap_script", "bootstrapScript"], default=None): t.Null | t.String,
        t.Key("dependencies", default=None): t.Null | t.List(tx.UUID) | t.List(t.String),
        tx.AliasedKey(["callback_url", "callbackUrl", "callbackURL"], default=None): (
            t.Null | tx.URL
        ),
        t.Key("owner_access_key", default=None): t.Null | t.String,
    }),
    loads=_json_loads,
)
async def create_from_params(request: web.Request, params: dict[str, Any]) -> web.Response:
    if params["session_name"] in ["from-template"]:
        raise InvalidAPIParameters(
            f"Requested session ID {params['session_name']} is reserved word"
        )
    api_version = request["api_version"]
    if 9 <= api_version[0]:
        creation_config = creation_config_v7.check(params["config"])
    elif 8 <= api_version[0]:
        creation_config = creation_config_v6.check(params["config"])
    elif 6 <= api_version[0]:
        creation_config = creation_config_v5.check(params["config"])
    elif 5 <= api_version[0]:
        creation_config = creation_config_v4.check(params["config"])
    elif (4, "20190315") <= api_version:
        creation_config = creation_config_v3.check(params["config"])
    elif 2 <= api_version[0] <= 4:
        creation_config = creation_config_v2.check(params["config"])
    elif api_version[0] == 1:
        creation_config = creation_config_v1.check(params["config"])
    else:
        raise InvalidAPIParameters("API version not supported")
    params["config"] = creation_config

    root_ctx: RootContext = request.app["_root.context"]

    agent_list = cast(Optional[list[str]], params["config"]["agent_list"])
    if agent_list is not None:
        if (
            request["user"]["role"] != UserRole.SUPERADMIN
            and root_ctx.config_provider.config.manager.hide_agents
        ):
            raise InsufficientPrivilege(
                "You are not allowed to manually assign agents for your session."
            )

    if params["domain"] is None:
        domain_name = request["user"]["domain_name"]
    else:
        domain_name = params["domain"]
    scopes_param = {
        "owner_access_key": (
            None if params["owner_access_key"] is undefined else params["owner_access_key"]
        ),
    }
    requester_access_key, owner_access_key = await get_access_key_scopes(request, scopes_param)
    log.info(
        "GET_OR_CREATE (ak:{0}/{1}, img:{2}, s:{3})",
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else "*",
        params["image"],
        params["session_name"],
    )

    result = await root_ctx.processors.session.create_from_params.wait_for_complete(
        CreateFromParamsAction(
            params=CreateFromParamsActionParams(
                session_name=params["session_name"],
                image=params["image"],
                architecture=params["architecture"],
                session_type=params["session_type"],
                group_name=params["group"],
                domain_name=domain_name,
                cluster_size=params["cluster_size"],
                cluster_mode=params["cluster_mode"],
                config=params["config"],
                tag=params["tag"],
                enqueue_only=params["enqueue_only"],
                max_wait_seconds=params["max_wait_seconds"],
                reuse_if_exists=params["reuse"],
                startup_command=params["startup_command"],
                bootstrap_script=params["bootstrap_script"],
                dependencies=params["dependencies"],
                callback_url=params["callback_url"],
                priority=params["priority"],
                starts_at=params["starts_at"],
                batch_timeout=params["batch_timeout"],
                owner_access_key=owner_access_key,
            ),
            user_id=request["user"]["uuid"],
            user_role=request["user"]["role"],
            requester_access_key=requester_access_key,
            sudo_session_enabled=request["user"]["sudo_session_enabled"],
            keypair_resource_policy=request["keypair"]["resource_policy"],
        )
    )

    return web.json_response(result.result, status=HTTPStatus.CREATED)


@server_status_required(ALL_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        t.Key("clientSessionToken") >> "session_name": tx.SessionName,
        tx.AliasedKey(["template_id", "templateId"]): t.Null | tx.UUID,
        tx.AliasedKey(["type", "sessionType"], default="interactive") >> "sess_type": tx.Enum(
            SessionTypes
        ),
        tx.AliasedKey(["group", "groupName", "group_name"], default="default"): t.String,
        tx.AliasedKey(["domain", "domainName", "domain_name"], default="default"): t.String,
        tx.AliasedKey(["scaling_group", "scalingGroup"], default=None): t.Null | t.String,
        t.Key("tag", default=None): t.Null | t.String,
        t.Key("enqueueOnly", default=False) >> "enqueue_only": t.ToBool,
        t.Key("maxWaitSeconds", default=0) >> "max_wait_seconds": t.Int[0:],
        t.Key("owner_access_key", default=None): t.Null | t.String,
    }),
    loads=_json_loads,
)
async def create_cluster(request: web.Request, params: dict[str, Any]) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    if params["domain"] is None:
        domain_name = request["user"]["domain_name"]
    else:
        domain_name = params["domain"]
    scopes_param = {
        "owner_access_key": (
            None if params["owner_access_key"] is undefined else params["owner_access_key"]
        ),
    }
    requester_access_key, owner_access_key = await get_access_key_scopes(request, scopes_param)
    log.info(
        "CREAT_CLUSTER (ak:{0}/{1}, s:{2})",
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else "*",
        params["session_name"],
    )

    result = await root_ctx.processors.session.create_cluster.wait_for_complete(
        CreateClusterAction(
            session_name=params["session_name"],
            user_id=request["user"]["uuid"],
            user_role=request["user"]["role"],
            domain_name=domain_name,
            group_name=params["group"],
            requester_access_key=requester_access_key,
            owner_access_key=owner_access_key,
            scaling_group_name=params["scaling_group"],
            tag=params["tag"],
            session_type=params["sess_type"],
            enqueue_only=params["enqueue_only"],
            template_id=params["template_id"],
            sudo_session_enabled=request["user"]["sudo_session_enabled"],
            max_wait_seconds=params["max_wait_seconds"],
            keypair_resource_policy=request["keypair"]["resource_policy"],
        )
    )

    return web.json_response(result.result, status=HTTPStatus.CREATED)


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        t.Key("login_session_token", default=None): t.Null | t.String,
        tx.AliasedKey(["app", "service"]): t.String,
        # The port argument is only required to use secondary ports
        # when the target app listens multiple TCP ports.
        # Otherwise it should be omitted or set to the same value of
        # the actual port number used by the app.
        tx.AliasedKey(["port"], default=None): t.Null | t.Int[1024:65535],
        tx.AliasedKey(["envs"], default=None): t.Null | t.String,  # stringified JSON
        # e.g., '{"PASSWORD": "12345"}'
        tx.AliasedKey(["arguments"], default=None): t.Null | t.String,  # stringified JSON
        # e.g., '{"-P": "12345"}'
        # The value can be one of:
        # None, str, List[str]
    })
)
async def start_service(request: web.Request, params: Mapping[str, Any]) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name: str = request.match_info["session_name"]
    access_key: AccessKey = request["keypair"]["access_key"]
    service: str = params["app"]
    myself = asyncio.current_task()
    if myself is None:
        raise NoCurrentTaskContext("No current task context")
    result = await root_ctx.processors.session.start_service.wait_for_complete(
        StartServiceAction(
            session_name=session_name,
            access_key=access_key,
            service=service,
            login_session_token=params.get("login_session_token"),
            port=params.get("port"),
            envs=params.get("envs"),
            arguments=params.get("arguments"),
        )
    )

    return web.json_response({
        "token": result.token,
        "wsproxy_addr": result.wsproxy_addr,
    })


@server_status_required(ALL_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        t.Key("login_session_token", default=None): t.Null | t.String,
    }),
    loads=_json_loads,
)
async def get_commit_status(request: web.Request, params: Mapping[str, Any]) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name: str = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)

    myself = asyncio.current_task()
    if myself is None:
        raise NoCurrentTaskContext("No current task context")

    log.info(
        "GET_COMMIT_STATUS (ak:{}/{}, s:{})", requester_access_key, owner_access_key, session_name
    )
    result = await root_ctx.processors.session.get_commit_status.wait_for_complete(
        GetCommitStatusAction(
            session_name=session_name,
            owner_access_key=owner_access_key,
        )
    )
    resp = result.commit_info.asdict()
    return web.json_response(resp, status=HTTPStatus.OK)


@server_status_required(ALL_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        t.Key("login_session_token", default=None): t.Null | t.String,
    }),
    loads=_json_loads,
)
async def get_abusing_report(request: web.Request, params: Mapping[str, Any]) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name: str = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)

    log.info(
        "GET_ABUSING_REPORT (ak:{}/{}, s:{})", requester_access_key, owner_access_key, session_name
    )
    result = await root_ctx.processors.session.get_abusing_report.wait_for_complete(
        GetAbusingReportAction(
            session_name=session_name,
            owner_access_key=owner_access_key,
        )
    )
    return web.json_response(result.abuse_report or {}, status=HTTPStatus.OK)


@server_status_required(ALL_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        t.Key("agent"): t.String,
    }),
)
async def sync_agent_registry(request: web.Request, params: Any) -> web.StreamResponse:
    root_ctx: RootContext = request.app["_root.context"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)

    agent_id = AgentId(params["agent"])
    log.info(
        "SYNC_AGENT_REGISTRY (ak:{}/{}, a:{})", requester_access_key, owner_access_key, agent_id
    )
    await root_ctx.processors.agent.sync_agent_registry.wait_for_complete(
        SyncAgentRegistryAction(
            agent_id=agent_id,
        )
    )
    return web.json_response({}, status=HTTPStatus.OK)


class TransitSessionStatusRequestModel(LegacyBaseRequestModel):
    ids: list[UUID] = Field(
        validation_alias=AliasChoices("session_ids", "sessionIds", "SessionIds"),
        description="ID array of sessions to check and transit status.",
    )


class SessionStatusResponseModel(LegacyBaseResponseModel):
    session_status_map: dict[SessionId, str]


@auth_required
@server_status_required(ALL_ALLOWED)
@pydantic_params_api_handler(TransitSessionStatusRequestModel)
async def check_and_transit_status(
    request: web.Request, params: TransitSessionStatusRequestModel
) -> SessionStatusResponseModel:
    root_ctx: RootContext = request.app["_root.context"]
    session_ids = [SessionId(id) for id in params.ids]
    user_role = cast(UserRole, request["user"]["role"])
    user_id = cast(UUID, request["user"]["uuid"])
    requester_access_key, owner_access_key = await get_access_key_scopes(request)
    log.info("TRANSIT_STATUS (ak:{}/{}, s:{})", requester_access_key, owner_access_key, session_ids)

    session_status_map = {}
    for session_id in session_ids:
        result = await root_ctx.processors.session.check_and_transit_status.wait_for_complete(
            CheckAndTransitStatusAction(
                user_id=user_id,
                user_role=user_role,
                session_id=session_id,
            )
        )
        session_status_map.update(result.result)
    return SessionStatusResponseModel(session_status_map=session_status_map)


@server_status_required(ALL_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        t.Key("login_session_token", default=None): t.Null | t.String,
        # if `dst` is None, it will be agent's default destination.
        tx.AliasedKey(["filename", "fname"], default=None): t.Null | t.String,
    }),
    loads=_json_loads,
)
async def commit_session(request: web.Request, params: Mapping[str, Any]) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name: str = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)
    filename: Optional[str] = params["filename"]

    log.info(
        "COMMIT_SESSION (ak:{}/{}, s:{})", requester_access_key, owner_access_key, session_name
    )

    action_result = await root_ctx.processors.session.commit_session.wait_for_complete(
        CommitSessionAction(
            session_name=session_name,
            owner_access_key=owner_access_key,
            filename=filename,
        )
    )

    return web.json_response(action_result.commit_result, status=HTTPStatus.CREATED)


class ConvertSessionToImageRequesteModel(LegacyBaseRequestModel):
    image_name: str = Field(
        pattern=r"[a-zA-Z0-9\.\-_]+",
        description="Name of the image to be created.",
    )
    login_session_token: Annotated[Optional[str], Field(default=None)]
    image_visibility: CustomizedImageVisibilityScope = Field(
        default=CustomizedImageVisibilityScope.USER,
        description="Visibility scope of newly created image. currently only supports `USER` scope. Setting this to value other than `USER` will raise error.",
    )


class ConvertSessionToImageResponseModel(LegacyBaseResponseModel):
    task_id: str


@auth_required
@server_status_required(ALL_ALLOWED)
@pydantic_params_api_handler(ConvertSessionToImageRequesteModel)
async def convert_session_to_image(
    request: web.Request, params: ConvertSessionToImageRequesteModel
) -> ConvertSessionToImageResponseModel:
    root_ctx: RootContext = request.app["_root.context"]
    session_name: str = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)

    log.info(
        "CONVERT_SESSION_TO_IMAGE (ak:{}/{}, s:{})",
        requester_access_key,
        owner_access_key,
        session_name,
    )

    result = await root_ctx.processors.session.convert_session_to_image.wait_for_complete(
        ConvertSessionToImageAction(
            session_name=session_name,
            owner_access_key=owner_access_key,
            image_name=params.image_name,
            image_visibility=params.image_visibility,
            image_owner_id=request["user"]["uuid"],
            user_email=request["user"]["email"],
            max_customized_image_count=request["user"]["resource_policy"][
                "max_customized_image_count"
            ],
        )
    )
    return ConvertSessionToImageResponseModel(task_id=str(result.task_id))


@catch_unexpected(log)
async def check_agent_lost(root_ctx: RootContext, interval: float) -> None:
    try:
        now = datetime.now(tzutc())
        timeout = timedelta(seconds=root_ctx.config_provider.config.manager.heartbeat_timeout)

        agent_last_seen = await root_ctx.valkey_live.scan_agent_last_seen()
        for agent_id, prev_timestamp in agent_last_seen:
            prev = datetime.fromtimestamp(prev_timestamp, tzutc())
            if now - prev > timeout:
                await root_ctx.event_producer.anycast_event(
                    AgentTerminatedEvent("agent-lost"),
                    source_override=AgentId(agent_id),
                )
    except asyncio.CancelledError:
        pass


@catch_unexpected(log)
async def report_stats(root_ctx: RootContext, interval: float) -> None:
    try:
        stats_monitor = root_ctx.stats_monitor
        await stats_monitor.report_metric(
            GAUGE, "ai.backend.manager.coroutines", len(asyncio.all_tasks())
        )

        all_inst_ids = [inst_id async for inst_id in root_ctx.registry.enumerate_instances()]
        await stats_monitor.report_metric(
            GAUGE, "ai.backend.manager.agent_instances", len(all_inst_ids)
        )

        async with root_ctx.db.begin_readonly() as conn:
            query = (
                sa.select([sa.func.count()])
                .select_from(kernels)
                .where(
                    (kernels.c.cluster_role == DEFAULT_ROLE)
                    & (kernels.c.status.in_(AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES)),
                )
            )
            n = await conn.scalar(query)
            await stats_monitor.report_metric(GAUGE, "ai.backend.manager.active_kernels", n)
            subquery = (
                sa.select([sa.func.count()])
                .select_from(keypairs)
                .where(keypairs.c.is_active == true())
                .group_by(keypairs.c.user_id)
            )
            query = sa.select([sa.func.count()]).select_from(subquery.alias())
            n = await conn.scalar(query)
            await stats_monitor.report_metric(GAUGE, "ai.backend.users.has_active_key", n)

            subquery = subquery.where(keypairs.c.last_used != null())
            query = sa.select([sa.func.count()]).select_from(subquery.alias())
            n = await conn.scalar(query)
            await stats_monitor.report_metric(GAUGE, "ai.backend.users.has_used_key", n)

            """
            query = sa.select([sa.func.count()]).select_from(usage)
            n = await conn.scalar(query)
            await stats_monitor.report_metric(
                GAUGE, 'ai.backend.manager.accum_kernels', n)
            """
    except (sqlalchemy.exc.InterfaceError, ConnectionRefusedError):
        log.warning("report_stats(): error while connecting to PostgreSQL server")


@server_status_required(ALL_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(["name", "session_name", "clientSessionToken"])
        >> "session_name": tx.SessionName,
    }),
)
async def rename_session(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name = request.match_info["session_name"]
    new_name = params["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)
    log.info(
        "RENAME_SESSION (ak:{0}/{1}, s:{2}, newname:{3})",
        request,
        owner_access_key,
        session_name,
        new_name,
    )

    await root_ctx.processors.session.rename_session.wait_for_complete(
        RenameSessionAction(
            session_name=session_name,
            new_name=new_name,
            owner_access_key=owner_access_key,
        )
    )

    return web.Response(status=HTTPStatus.NO_CONTENT)


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        t.Key("forced", default="false"): t.ToBool(),
        t.Key("recursive", default="false"): t.ToBool(),
        t.Key("owner_access_key", default=None): t.Null | t.String,
    })
)
async def destroy(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name = request.match_info["session_name"]
    user_role = cast(UserRole, request["user"]["role"])
    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    if requester_access_key != owner_access_key and user_role not in (
        UserRole.ADMIN,
        UserRole.SUPERADMIN,
    ):
        raise InsufficientPrivilege("You are not allowed to force-terminate others's sessions")

    log.info(
        "DESTROY (ak:{0}/{1}, s:{2}, forced:{3}, recursive: {4})",
        requester_access_key,
        owner_access_key,
        session_name,
        params["forced"],
        params["recursive"],
    )

    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)

    result = await root_ctx.processors.session.destroy_session.wait_for_complete(
        DestroySessionAction(
            session_name=session_name,
            owner_access_key=owner_access_key,
            user_role=user_role,
            forced=params["forced"],
            recursive=params["recursive"],
        )
    )
    return web.json_response(result.result, status=HTTPStatus.OK)


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        t.Key("id"): t.String(),
    })
)
async def match_sessions(request: web.Request, params: Any) -> web.Response:
    """
    A quick session-ID matcher API for use with auto-completion in CLI.
    """
    root_ctx: RootContext = request.app["_root.context"]
    id_or_name_prefix = params["id"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)
    log.info(
        "MATCH_SESSIONS(ak:{0}/{1}, prefix:{2})",
        requester_access_key,
        owner_access_key,
        id_or_name_prefix,
    )
    result = await root_ctx.processors.session.match_sessions.wait_for_complete(
        MatchSessionsAction(
            id_or_name_prefix=id_or_name_prefix,
            owner_access_key=owner_access_key,
        )
    )

    return web.json_response(
        {
            "matches": result.result,
        },
        status=HTTPStatus.OK,
    )


@server_status_required(READ_ALLOWED)
@auth_required
async def get_direct_access_info(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name = request.match_info["session_name"]
    _, owner_access_key = await get_access_key_scopes(request)

    result = await root_ctx.processors.session.get_direct_access_info.wait_for_complete(
        GetDirectAccessInfoAction(
            session_name=session_name,
            owner_access_key=owner_access_key,
        )
    )
    return web.json_response(result.result)


@server_status_required(READ_ALLOWED)
@auth_required
async def get_info(request: web.Request) -> web.Response:
    # NOTE: This API should be replaced with GraphQL version.
    root_ctx: RootContext = request.app["_root.context"]
    session_name = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)
    log.info("GET_INFO (ak:{0}/{1}, s:{2})", requester_access_key, owner_access_key, session_name)
    try:
        result = await root_ctx.processors.session.get_session_info.wait_for_complete(
            GetSessionInfoAction(
                session_name=session_name,
                owner_access_key=owner_access_key,
            )
        )
    except BackendAIError:
        log.exception("GET_INFO: exception")
        raise
    return web.json_response(result.session_info.asdict(), status=HTTPStatus.OK)


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        t.Key("owner_access_key", default=None): t.Null | t.String,
    })
)
async def restart(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info("RESTART (ak:{0}/{1}, s:{2})", requester_access_key, owner_access_key, session_name)

    try:
        await root_ctx.processors.session.restart_session.wait_for_complete(
            RestartSessionAction(
                session_name=session_name,
                owner_access_key=owner_access_key,
            )
        )
    except BackendAIError:
        log.exception("RESTART: exception")
        raise
    except Exception:
        await root_ctx.error_monitor.capture_exception(context={"user": request["user"]["uuid"]})
        log.exception("RESTART: unexpected error")
        raise web.HTTPInternalServerError
    return web.Response(status=HTTPStatus.NO_CONTENT)


@server_status_required(READ_ALLOWED)
@auth_required
async def execute(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)
    try:
        params = await read_json(request)
        log.info("EXECUTE(ak:{0}/{1}, s:{2})", requester_access_key, owner_access_key, session_name)
    except json.decoder.JSONDecodeError:
        log.warning("EXECUTE: invalid/missing parameters")
        raise InvalidAPIParameters

    result = await root_ctx.processors.session.execute_session.wait_for_complete(
        ExecuteSessionAction(
            session_name=session_name,
            owner_access_key=owner_access_key,
            api_version=request["api_version"],
            params=ExecuteSessionActionParams(
                mode=params.get("mode", None),
                run_id=params.get("run_id", None),
                code=params.get("code", None),
                options=params.get("options", None),
            ),
        )
    )
    return web.json_response(result.result, status=HTTPStatus.OK)


@server_status_required(READ_ALLOWED)
@auth_required
async def interrupt(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)
    log.info("INTERRUPT(ak:{0}/{1}, s:{2})", requester_access_key, owner_access_key, session_name)

    try:
        await root_ctx.processors.session.interrupt.wait_for_complete(
            InterruptSessionAction(
                session_name=session_name,
                owner_access_key=owner_access_key,
            )
        )
    except BackendAIError:
        log.exception("INTERRUPT: exception")
        raise
    return web.Response(status=HTTPStatus.NO_CONTENT)


@server_status_required(READ_ALLOWED)
@auth_required
async def complete(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)

    try:
        params = await read_json(request)
        code = params.get("code", "")
        opts = params.get("options", None) or {}
    except json.decoder.JSONDecodeError:
        raise InvalidAPIParameters

    log.info("COMPLETE(ak:{0}/{1}, s:{2})", requester_access_key, owner_access_key, session_name)

    action_result = await root_ctx.processors.session.complete.wait_for_complete(
        CompleteAction(
            session_name=session_name,
            owner_access_key=owner_access_key,
            code=code,
            options=opts,
        )
    )

    return web.json_response(action_result.result.as_dict(), status=HTTPStatus.OK)


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        t.Key("service_name"): t.String,
    })
)
async def shutdown_service(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)
    log.info(
        "SHUTDOWN_SERVICE (ak:{0}/{1}, s:{2})", requester_access_key, owner_access_key, session_name
    )
    service_name = params.get("service_name")

    try:
        await root_ctx.processors.session.shutdown_service.wait_for_complete(
            ShutdownServiceAction(
                session_name=session_name,
                owner_access_key=owner_access_key,
                service_name=service_name,
            )
        )
    except BackendAIError:
        log.exception("SHUTDOWN_SERVICE: exception")
        raise
    return web.Response(status=HTTPStatus.NO_CONTENT)


async def find_dependent_sessions(
    root_session_name_or_id: str | UUID,
    db_session: SASession,
    access_key: AccessKey,
    *,
    allow_stale: bool = False,
) -> set[UUID]:
    async def _find_dependent_sessions(session_id: UUID) -> set[UUID]:
        result = await db_session.execute(
            sa.select(SessionDependencyRow).where(SessionDependencyRow.depends_on == session_id)
        )
        dependent_sessions: set[UUID] = {x.session_id for x in result.scalars()}

        recursive_dependent_sessions: list[set[UUID]] = [
            await _find_dependent_sessions(dependent_session)
            for dependent_session in dependent_sessions
        ]

        for recursive_dependent_session in recursive_dependent_sessions:
            dependent_sessions |= recursive_dependent_session

        return dependent_sessions

    root_session = await SessionRow.get_session(
        db_session,
        root_session_name_or_id,
        access_key=access_key,
        allow_stale=allow_stale,
    )
    return await _find_dependent_sessions(cast(UUID, root_session.id))


@aiotools.lru_cache(maxsize=100)
async def _find_dependency_sessions(
    session_name_or_id: UUID | str,
    db_session: SASession,
    access_key: AccessKey,
):
    sessions = await SessionRow.match_sessions(
        db_session,
        session_name_or_id,
        access_key=access_key,
    )

    if len(sessions) < 1:
        raise SessionNotFound("session not found!")

    session_id = str(sessions[0].id)
    session_name = sessions[0].name

    if not isinstance(session_name, str):
        raise InvalidSessionData("Invalid session_name type")

    kernel_query = (
        sa.select([
            kernels.c.status,
            kernels.c.status_changed,
        ])
        .select_from(kernels)
        .where(kernels.c.session_id == session_id)
    )

    dependency_session_ids: list[SessionDependencyRow] = (
        await db_session.execute(
            sa.select(SessionDependencyRow.depends_on).where(
                SessionDependencyRow.session_id == session_id
            )
        )
    ).first()

    if not dependency_session_ids:
        dependency_session_ids = []

    kernel_query_result = (await db_session.execute(kernel_query)).first()

    session_info: dict[str, list | str] = {
        "session_id": session_id,
        "session_name": session_name,
        "status": str(kernel_query_result[0]),
        "status_changed": str(kernel_query_result[1]),
        "depends_on": [
            await _find_dependency_sessions(dependency_session_id, db_session, access_key)
            for dependency_session_id in dependency_session_ids
        ],
    }

    return session_info


async def find_dependency_sessions(
    session_name_or_id: UUID | str,
    db_session: SASession,
    access_key: AccessKey,
):
    return await _find_dependency_sessions(session_name_or_id, db_session, access_key)


@server_status_required(READ_ALLOWED)
@auth_required
async def get_dependency_graph(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    root_session_name = request.match_info["session_name"]

    requester_access_key, owner_access_key = await get_access_key_scopes(request)

    log.info(
        "GET_DEPENDENCY_GRAPH (ak:{0}/{1}, s:{2})",
        requester_access_key,
        owner_access_key,
        root_session_name,
    )

    result = await root_ctx.processors.session.get_dependency_graph.wait_for_complete(
        GetDependencyGraphAction(
            root_session_name=root_session_name,
            owner_access_key=owner_access_key,
        )
    )

    return web.json_response(
        result.result,
        status=HTTPStatus.OK,
    )


@server_status_required(READ_ALLOWED)
@auth_required
async def upload_files(request: web.Request) -> web.Response:
    reader = await request.multipart()
    root_ctx: RootContext = request.app["_root.context"]
    session_name = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)
    log.info(
        "UPLOAD_FILE (ak:{0}/{1}, s:{2})", requester_access_key, owner_access_key, session_name
    )

    try:
        await root_ctx.processors.session.upload_files.wait_for_complete(
            UploadFilesAction(
                session_name=session_name,
                owner_access_key=owner_access_key,
                reader=reader,
            )
        )
    except BackendAIError:
        log.exception("UPLOAD_FILES: exception")
        raise
    return web.Response(status=HTTPStatus.NO_CONTENT)


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        tx.MultiKey("files"): t.List(t.String),
    })
)
async def download_files(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)
    files = params.get("files")
    log.info(
        "DOWNLOAD_FILE (ak:{0}/{1}, s:{2}, path:{3!r})",
        requester_access_key,
        owner_access_key,
        session_name,
        files[0],
    )
    result = await root_ctx.processors.session.download_files.wait_for_complete(
        DownloadFilesAction(
            user_id=request["user"]["uuid"],
            owner_access_key=owner_access_key,
            session_name=session_name,
            files=files,
        )
    )
    return web.Response(body=result.result, status=HTTPStatus.OK)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict({
        t.Key("file"): t.String,
    })
)
async def download_single(request: web.Request, params: Any) -> web.Response:
    """
    Download a single file from the scratch root. Only for small files.
    """
    root_ctx: RootContext = request.app["_root.context"]
    session_name = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request)
    file = params["file"]
    log.info(
        "DOWNLOAD_SINGLE (ak:{0}/{1}, s:{2}, path:{3!r})",
        requester_access_key,
        owner_access_key,
        session_name,
        file,
    )

    result = await root_ctx.processors.session.download_file.wait_for_complete(
        DownloadFileAction(
            user_id=request["user"]["uuid"],
            session_name=session_name,
            owner_access_key=owner_access_key,
            file=file,
        )
    )
    return web.Response(body=result.bytes, status=HTTPStatus.OK)


@server_status_required(READ_ALLOWED)
@auth_required
async def list_files(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    try:
        session_name = request.match_info["session_name"]
        requester_access_key, owner_access_key = await get_access_key_scopes(request)
        params = await read_json(request)
        path = params.get("path", ".")
        log.info(
            "LIST_FILES (ak:{0}/{1}, s:{2}, path:{3})",
            requester_access_key,
            owner_access_key,
            session_name,
            path,
        )
        result = await root_ctx.processors.session.list_files.wait_for_complete(
            ListFilesAction(
                user_id=request["user"]["uuid"],
                path=path,
                session_name=session_name,
                owner_access_key=owner_access_key,
            )
        )
    except (asyncio.TimeoutError, AssertionError, json.decoder.JSONDecodeError) as e:
        log.warning("LIST_FILES: invalid/missing parameters, {0!r}", e)
        raise InvalidAPIParameters(extra_msg=str(e.args[0]))

    return web.json_response(result.result, status=HTTPStatus.OK)


class ContainerLogRequestModel(LegacyBaseRequestModel):
    owner_access_key: Optional[str] = Field(
        default=None,
        alias="ownerAccessKey",
    )
    kernel_id: Optional[UUID] = Field(
        description="Target kernel to get container logs.",
        default=None,
        alias="kernelId",
    )


@server_status_required(READ_ALLOWED)
@auth_required
@pydantic_params_api_handler(ContainerLogRequestModel)
async def get_container_logs(
    request: web.Request, params: ContainerLogRequestModel
) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name: str = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(
        request, {"owner_access_key": params.owner_access_key}
    )
    # assume retrieving container log of main kernel when `params.kernel_id` is None
    kernel_id = KernelId(params.kernel_id) if params.kernel_id is not None else None
    log.info(
        "GET_CONTAINER_LOG (ak:{}/{}, s:{}, k:{})",
        requester_access_key,
        owner_access_key,
        session_name,
        kernel_id,
    )

    try:
        result = await root_ctx.processors.session.get_container_logs.wait_for_complete(
            GetContainerLogsAction(
                session_name=session_name,
                owner_access_key=owner_access_key,
                kernel_id=kernel_id,
            )
        )
    except BackendAIError:
        log.exception(
            "GET_CONTAINER_LOG(ak:{}/{}, kernel_id: {}, s:{}): unexpected error",
            requester_access_key,
            owner_access_key,
            kernel_id,
            session_name,
        )
        raise
    return web.json_response(result.result, status=HTTPStatus.OK)


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(["session_name", "sessionName", "task_id", "taskId"]) >> "kernel_id": tx.UUID,
    })
)
async def get_task_logs(request: web.Request, params: Any) -> web.StreamResponse:
    log.info("GET_TASK_LOG (ak:{}, k:{})", request["keypair"]["access_key"], params["kernel_id"])
    root_ctx: RootContext = request.app["_root.context"]
    domain_name = request["user"]["domain_name"]
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]
    kernel_id_str = KernelId(params["kernel_id"])

    result = await root_ctx.processors.vfolder.get_task_logs.wait_for_complete(
        GetTaskLogsAction(
            user_id=user_uuid,
            domain_name=domain_name,
            user_role=user_role,
            kernel_id=KernelId(kernel_id_str),
            owner_access_key=request["keypair"]["access_key"],
            request=request,
        )
    )
    return result.response


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        t.Key("owner_access_key", default=None): t.Null | t.String,
    })
)
async def get_status_history(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    session_name: str = request.match_info["session_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info(
        "GET_STATUS_HISTORY (ak:{}/{}, s:{})", requester_access_key, owner_access_key, session_name
    )

    resp: dict[str, Mapping] = {"result": {}}
    result = await root_ctx.processors.session.get_status_history.wait_for_complete(
        GetStatusHistoryAction(
            session_name=session_name,
            owner_access_key=request["keypair"]["access_key"],
        )
    )

    resp["result"] = result.status_history
    return web.json_response(resp, status=200)


@attrs.define(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    agent_lost_checker: asyncio.Task[None]
    stats_task: asyncio.Task[None]
    database_ptask_group: aiotools.PersistentTaskGroup
    rpc_ptask_group: aiotools.PersistentTaskGroup
    webhook_ptask_group: aiotools.PersistentTaskGroup


async def init(app: web.Application) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["session.context"]
    app_ctx.database_ptask_group = aiotools.PersistentTaskGroup()
    app_ctx.rpc_ptask_group = aiotools.PersistentTaskGroup()
    app_ctx.webhook_ptask_group = aiotools.PersistentTaskGroup()

    # Scan ALIVE agents
    app_ctx.agent_lost_checker = aiotools.create_timer(
        functools.partial(check_agent_lost, root_ctx), 1.0
    )
    app_ctx.agent_lost_checker.set_name("agent_lost_checker")
    app_ctx.stats_task = aiotools.create_timer(
        functools.partial(report_stats, root_ctx),
        5.0,
    )
    app_ctx.stats_task.set_name("stats_task")


async def shutdown(app: web.Application) -> None:
    app_ctx: PrivateContext = app["session.context"]
    await cancel_and_wait(app_ctx.agent_lost_checker)
    await cancel_and_wait(app_ctx.stats_task)

    await app_ctx.webhook_ptask_group.shutdown()
    await app_ctx.database_ptask_group.shutdown()
    await app_ctx.rpc_ptask_group.shutdown()


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app["api_versions"] = (1, 2, 3, 4)
    app["session.context"] = PrivateContext()
    app["prefix"] = "session"
    deprecated_get_stub = deprecated_stub(
        "Use the HTTP POST method to invoke this API with parameters in the request body."
    )
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route("POST", "", create_from_params))
    cors.add(app.router.add_route("POST", "/_/create", create_from_params))
    cors.add(app.router.add_route("POST", "/_/create-from-template", create_from_template))
    cors.add(app.router.add_route("POST", "/_/create-cluster", create_cluster))
    cors.add(app.router.add_route("GET", "/_/match", match_sessions))
    cors.add(app.router.add_route("POST", "/_/sync-agent-registry", sync_agent_registry))
    cors.add(app.router.add_route("POST", "/_/transit-status", check_and_transit_status))
    session_resource = cors.add(app.router.add_resource(r"/{session_name}"))
    cors.add(session_resource.add_route("GET", get_info))
    cors.add(session_resource.add_route("PATCH", restart))
    cors.add(session_resource.add_route("DELETE", destroy))
    cors.add(session_resource.add_route("POST", execute))
    task_log_resource = cors.add(app.router.add_resource(r"/_/logs"))
    cors.add(task_log_resource.add_route("HEAD", get_task_logs))
    cors.add(task_log_resource.add_route("GET", get_task_logs))
    cors.add(
        app.router.add_route("GET", "/{session_name}/direct-access-info", get_direct_access_info)
    )
    cors.add(app.router.add_route("GET", "/{session_name}/logs", get_container_logs))
    cors.add(app.router.add_route("POST", "/{session_name}/rename", rename_session))
    cors.add(app.router.add_route("POST", "/{session_name}/interrupt", interrupt))
    cors.add(app.router.add_route("POST", "/{session_name}/complete", complete))
    cors.add(app.router.add_route("POST", "/{session_name}/shutdown-service", shutdown_service))
    cors.add(app.router.add_route("POST", "/{session_name}/upload", upload_files))
    cors.add(app.router.add_route("GET", "/{session_name}/download", deprecated_get_stub))
    cors.add(app.router.add_route("GET", "/{session_name}/download_single", deprecated_get_stub))
    cors.add(app.router.add_route("POST", "/{session_name}/download", download_files))
    cors.add(app.router.add_route("POST", "/{session_name}/download_single", download_single))
    cors.add(app.router.add_route("GET", "/{session_name}/files", list_files))
    cors.add(app.router.add_route("POST", "/{session_name}/start-service", start_service))
    cors.add(app.router.add_route("POST", "/{session_name}/commit", commit_session))
    cors.add(app.router.add_route("POST", "/{session_name}/imagify", convert_session_to_image))
    cors.add(app.router.add_route("GET", "/{session_name}/commit", get_commit_status))
    cors.add(app.router.add_route("GET", "/{session_name}/status-history", get_status_history))
    cors.add(app.router.add_route("GET", "/{session_name}/abusing-report", get_abusing_report))
    cors.add(app.router.add_route("GET", "/{session_name}/dependency-graph", get_dependency_graph))
    return app, []



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.huggingface_registry.types import HuggingFaceRegistryData
from ai.backend.manager.services.artifact_registry.actions.base import ArtifactRegistryAction


@dataclass
class GetHuggingFaceRegistriesAction(ArtifactRegistryAction):
    registry_ids: list[uuid.UUID]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "get_huggingface_registry_multi"


@dataclass
class GetHuggingFaceRegistriesActionResult(BaseActionResult):
    result: list[HuggingFaceRegistryData]

    @override
    def entity_id(self) -> Optional[str]:
        return None



"""Scheduling controller for managing session lifecycle and scheduling operations."""

from .scheduling_controller import SchedulingController, SchedulingControllerArgs

__all__ = [
    "SchedulingController",
    "SchedulingControllerArgs",
]



"""add `groups.type`

Revision ID: 308bcecec5c2
Revises: 8c74e7df26f8
Create Date: 2023-11-30 14:18:59.565099

"""

import enum
from typing import cast
from uuid import UUID, uuid4

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.sql import text

from ai.backend.manager.models.base import GUID, EnumValueType, IDColumn, convention

# revision identifiers, used by Alembic.
revision = "308bcecec5c2"
down_revision = "8c74e7df26f8"
branch_labels = None
depends_on = None

metadata = sa.MetaData(naming_convention=convention)
MAXIMUM_DOTFILE_SIZE = 64 * 1024  # 61 KiB
DEFAULT_PROJECT_RESOURCE_POLICY_NAME = "default"


class ProjectType(enum.StrEnum):
    GENERAL = "general"
    MODEL_STORE = "model-store"


users = sa.Table(
    "users",
    metadata,
    IDColumn("uuid"),
    sa.Column("domain_name", sa.String(length=64), sa.ForeignKey("domains.name"), index=True),
)
groups = sa.Table(
    "groups",
    metadata,
    IDColumn("id"),
    sa.Column("name", sa.String(length=64), nullable=False),
    sa.Column("description", sa.String(length=512)),
    sa.Column("is_active", sa.Boolean, default=True),
    sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
    sa.Column(
        "modified_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        onupdate=sa.func.current_timestamp(),
    ),
    #: Field for synchronization with external services.
    sa.Column("integration_id", sa.String(length=512)),
    sa.Column(
        "domain_name",
        sa.String(length=64),
        sa.ForeignKey("domains.name", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
        index=True,
    ),
    # TODO: separate resource-related fields with new domain resource policy table when needed.
    sa.Column("total_resource_slots", JSONB(), default={}),
    sa.Column(
        "allowed_vfolder_hosts",
        JSONB(),
        nullable=False,
        default={},
    ),
    # dotfiles column, \x90 means empty list in msgpack
    sa.Column(
        "dotfiles", sa.LargeBinary(length=MAXIMUM_DOTFILE_SIZE), nullable=False, default=b"\x90"
    ),
    sa.Column(
        "resource_policy",
        sa.String(length=256),
        sa.ForeignKey("project_resource_policies.name"),
        nullable=False,
    ),
    sa.Column(
        "type",
        EnumValueType(ProjectType),
        nullable=False,
        default=ProjectType.GENERAL,
    ),
    sa.UniqueConstraint("name", "domain_name", name="uq_groups_name_domain_name"),
)
association_groups_users = sa.Table(
    "association_groups_users",
    metadata,
    sa.Column(
        "user_id",
        GUID,
        sa.ForeignKey("users.uuid", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
        primary_key=True,
    ),
    sa.Column(
        "group_id",
        GUID,
        sa.ForeignKey("groups.id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
        primary_key=True,
    ),
)


projecttype_choices = list(map(str, ProjectType))
projecttype = postgresql.ENUM(
    *projecttype_choices,
    name="projecttype",
)


def upgrade():
    conn = op.get_bind()
    projecttype.create(conn)
    op.add_column(
        "groups",
        sa.Column("type", sa.Enum(*projecttype_choices, name="projecttype"), default="general"),
    )
    conn.execute(text("UPDATE groups SET type = 'general'"))
    op.alter_column("groups", "type", nullable=False)

    domain_names = conn.scalars(text("SELECT name FROM domains")).all()
    domain_names = cast(list[str], domain_names)
    project_resource_policies = conn.scalars(
        text("SELECT name FROM project_resource_policies")
    ).all()
    resource_policy_names = cast(list[str], project_resource_policies)
    try:
        picked_resource_policy = resource_policy_names[0]
    except IndexError:
        conn.execute(
            text(
                f"""INSERT INTO project_resource_policies
                (name, max_vfolder_count, max_quota_scope_size, max_network_count)
                VALUES ('{DEFAULT_PROJECT_RESOURCE_POLICY_NAME}', 0, -1, 3)"""
            )
        )
        picked_resource_policy = DEFAULT_PROJECT_RESOURCE_POLICY_NAME
    for domain_name in domain_names:
        model_store_gid = uuid4()
        conn.execute(
            sa.insert(groups).values({
                "id": model_store_gid,
                "name": "model-store",
                "domain_name": domain_name,
                "resource_policy": picked_resource_policy,
                "type": ProjectType.MODEL_STORE,
            })
        )

        uids = conn.scalars(sa.select(users.c.uuid).where(users.c.domain_name == domain_name)).all()
        uids = cast(list[UUID], uids)
        if uids:
            conn.execute(
                sa.insert(association_groups_users).values(
                    user_id=sa.bindparam("user_id"),
                    group_id=model_store_gid,
                ),
                [{"user_id": uid} for uid in uids],
            )


def downgrade():
    conn = op.get_bind()
    cmd = """DELETE FROM groups WHERE "type" = 'model-store'"""
    conn.execute(text(cmd))
    op.drop_column("groups", "type")
    projecttype.drop(conn)



from typing import Any, Mapping

import sqlalchemy as sa

from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.resource.types import ProjectResourcePolicyData
from ai.backend.manager.errors.common import ObjectNotFound
from ai.backend.manager.models.resource_policy import ProjectResourcePolicyRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

project_resource_policy_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(
                domain=DomainType.REPOSITORY, layer=LayerType.PROJECT_RESOURCE_POLICY_REPOSITORY
            )
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class ProjectResourcePolicyRepository:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    @project_resource_policy_repository_resilience.apply()
    async def create(self, fields: Mapping[str, Any]) -> ProjectResourcePolicyData:
        async with self._db.begin_session() as db_sess:
            db_row = ProjectResourcePolicyRow(**fields)
            db_sess.add(db_row)
            await db_sess.flush()
            return db_row.to_dataclass()

    @project_resource_policy_repository_resilience.apply()
    async def get_by_name(self, name: str) -> ProjectResourcePolicyData:
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(ProjectResourcePolicyRow).where(ProjectResourcePolicyRow.name == name)
            result = await db_sess.execute(query)
            row = result.scalar_one_or_none()
            if row is None:
                raise ObjectNotFound(f"Project resource policy with name {name} not found.")
            return row.to_dataclass()

    @project_resource_policy_repository_resilience.apply()
    async def update(self, name: str, fields: Mapping[str, Any]) -> ProjectResourcePolicyData:
        async with self._db.begin_session() as db_sess:
            query = sa.select(ProjectResourcePolicyRow).where(ProjectResourcePolicyRow.name == name)
            result = await db_sess.execute(query)
            row = result.scalar_one_or_none()
            if row is None:
                raise ObjectNotFound(f"Project resource policy with name {name} not found.")
            for key, value in fields.items():
                setattr(row, key, value)
            await db_sess.flush()
            return row.to_dataclass()

    @project_resource_policy_repository_resilience.apply()
    async def delete(self, name: str) -> ProjectResourcePolicyData:
        async with self._db.begin_session() as db_sess:
            query = sa.select(ProjectResourcePolicyRow).where(ProjectResourcePolicyRow.name == name)
            result = await db_sess.execute(query)
            row = result.scalar_one_or_none()
            if row is None:
                raise ObjectNotFound(f"Project resource policy with name {name} not found.")
            await db_sess.delete(row)
            return row.to_dataclass()



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.object_storage.modifier import ObjectStorageModifier
from ai.backend.manager.data.object_storage.types import ObjectStorageData
from ai.backend.manager.services.object_storage.actions.base import ObjectStorageAction


@dataclass
class UpdateObjectStorageAction(ObjectStorageAction):
    id: uuid.UUID
    modifier: ObjectStorageModifier

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "update"


@dataclass
class UpdateObjectStorageActionResult(BaseActionResult):
    result: ObjectStorageData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.result.id)



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.model_serving.types import RequesterCtx
from ai.backend.manager.services.model_serving.actions.base import ModelServiceAction


@dataclass
class DeleteRouteAction(ModelServiceAction):
    requester_ctx: RequesterCtx
    service_id: uuid.UUID
    route_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "delete"


@dataclass
class DeleteRouteActionResult(BaseActionResult):
    success: bool

    @override
    def entity_id(self) -> Optional[str]:
        return None



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.user.actions.base import UserAction


@dataclass
class DeleteUserAction(UserAction):
    email: str

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "delete"


@dataclass
class DeleteUserActionResult(BaseActionResult):
    @override
    def entity_id(self) -> Optional[str]:
        return None



"""
Concentrated agent selector implementation for sokovan scheduler.

This selector prefers agents with fewer available resources to maximize
resource utilization by concentrating workloads.
"""

from __future__ import annotations

import sys
from decimal import Decimal
from typing import Sequence, Union

from ai.backend.common.types import SessionTypes

from .selector import (
    AbstractAgentSelector,
    AgentSelectionConfig,
    AgentSelectionCriteria,
    AgentStateTracker,
    ResourceRequirements,
)
from .utils import count_unutilized_capabilities, order_slots_by_priority


class ConcentratedAgentSelector(AbstractAgentSelector):
    """
    Concentrated agent selector that maximizes resource utilization.

    This selector prefers agents with:
    1. Fewer kernels at the same endpoint (for endpoint replica spreading)
    2. Fewer unutilized capabilities
    3. Less available resources (to concentrate workloads)
    """

    def __init__(self, agent_selection_resource_priority: list[str]) -> None:
        """
        Initialize the concentrated selector.

        Args:
            agent_selection_resource_priority: Resource types in priority order
        """
        self.agent_selection_resource_priority = agent_selection_resource_priority

    def name(self) -> str:
        """
        Return the selector name for predicates.
        """
        return "ConcentratedAgentSelector"

    def success_message(self) -> str:
        """
        Return a message describing successful agent selection.
        """
        return "Agent selected using concentrated strategy for maximum resource utilization"

    def select_tracker_by_strategy(
        self,
        trackers: Sequence[AgentStateTracker],
        resource_req: ResourceRequirements,
        criteria: AgentSelectionCriteria,
        config: AgentSelectionConfig,
    ) -> AgentStateTracker:
        """
        Select an agent tracker to concentrate workloads.

        Assumes trackers are already filtered for compatibility.
        """
        # Sort requested slots by priority
        resource_priorities = order_slots_by_priority(
            resource_req.requested_slots, self.agent_selection_resource_priority
        )

        # Choose the tracker with minimum resources (to concentrate workloads)
        def tracker_sort_key(tracker: AgentStateTracker) -> tuple[Union[int, Decimal], ...]:
            agent = tracker.original_agent
            occupied_slots = tracker.get_current_occupied_slots()
            sort_key: list[int | Decimal] = []

            # First, consider kernel counts at endpoint for replica spreading
            if (
                config.enforce_spreading_endpoint_replica
                and criteria.kernel_counts_at_endpoint
                and criteria.session_metadata.session_type == SessionTypes.INFERENCE
            ):
                kernel_count = criteria.kernel_counts_at_endpoint.get(agent.agent_id, 0)
                sort_key.append(kernel_count)

            # Then, prefer agents with fewer unutilized capabilities
            sort_key.append(count_unutilized_capabilities(agent, resource_req.requested_slots))

            # Finally, prefer agents with less available resources (using current state)
            for key in resource_priorities:
                sort_key.append((agent.available_slots - occupied_slots).get(key, sys.maxsize))

            return tuple(sort_key)

        chosen_tracker = min(trackers, key=tracker_sort_key)

        return chosen_tracker



from dataclasses import dataclass
from typing import Optional, override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.deployment.modifier import NewDeploymentModifier
from ai.backend.manager.data.deployment.types import ModelDeploymentData
from ai.backend.manager.services.deployment.actions.base import DeploymentBaseAction


@dataclass
class UpdateDeploymentAction(DeploymentBaseAction):
    """Action to update an existing deployment."""

    deployment_id: UUID
    modifier: NewDeploymentModifier

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.deployment_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "update"


@dataclass
class UpdateDeploymentActionResult(BaseActionResult):
    data: ModelDeploymentData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.data.id)



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, Optional, override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult

from .base import NotificationAction


@dataclass
class ValidateRuleAction(NotificationAction):
    """Action to validate a notification rule by rendering its template with test data."""

    rule_id: UUID
    notification_data: Mapping[str, Any]

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "validate_rule"

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.rule_id)


@dataclass
class ValidateRuleActionResult(BaseActionResult):
    """Result of validating a notification rule."""

    message: str

    @override
    def entity_id(self) -> Optional[str]:
        return None



from __future__ import annotations

import uuid

import sqlalchemy as sa

from ai.backend.manager.data.notification import NotificationChannelType, NotificationRuleType
from ai.backend.manager.models.notification import NotificationChannelRow, NotificationRuleRow
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder


class NotificationChannelConditions:
    """Query conditions for notification channels."""

    @staticmethod
    def by_name_contains(name: str, case_insensitive: bool = False) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if case_insensitive:
                return NotificationChannelRow.name.ilike(f"%{name}%")
            else:
                return NotificationChannelRow.name.like(f"%{name}%")

        return inner

    @staticmethod
    def by_name_equals(name: str, case_insensitive: bool = False) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if case_insensitive:
                return sa.func.lower(NotificationChannelRow.name) == name.lower()
            else:
                return NotificationChannelRow.name == name

        return inner

    @staticmethod
    def by_channel_types(channel_types: list[NotificationChannelType]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return NotificationChannelRow.channel_type.in_(channel_types)

        return inner

    @staticmethod
    def by_enabled(enabled: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return NotificationChannelRow.enabled == enabled

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(NotificationChannelRow.created_at)
                .where(NotificationChannelRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return NotificationChannelRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(NotificationChannelRow.created_at)
                .where(NotificationChannelRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return NotificationChannelRow.created_at > subquery

        return inner


class NotificationChannelOrders:
    """Query orders for notification channels."""

    @staticmethod
    def name(ascending: bool = True) -> QueryOrder:
        if ascending:
            return NotificationChannelRow.name.asc()
        else:
            return NotificationChannelRow.name.desc()

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return NotificationChannelRow.created_at.asc()
        else:
            return NotificationChannelRow.created_at.desc()

    @staticmethod
    def updated_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return NotificationChannelRow.updated_at.asc()
        else:
            return NotificationChannelRow.updated_at.desc()


class NotificationRuleConditions:
    """Query conditions for notification rules."""

    @staticmethod
    def by_name_contains(name: str, case_insensitive: bool = False) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if case_insensitive:
                return NotificationRuleRow.name.ilike(f"%{name}%")
            else:
                return NotificationRuleRow.name.like(f"%{name}%")

        return inner

    @staticmethod
    def by_name_equals(name: str, case_insensitive: bool = False) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if case_insensitive:
                return sa.func.lower(NotificationRuleRow.name) == name.lower()
            else:
                return NotificationRuleRow.name == name

        return inner

    @staticmethod
    def by_rule_types(rule_types: list[NotificationRuleType]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return NotificationRuleRow.rule_type.in_(rule_types)

        return inner

    @staticmethod
    def by_enabled(enabled: bool) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return NotificationRuleRow.enabled == enabled

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(NotificationRuleRow.created_at)
                .where(NotificationRuleRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return NotificationRuleRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(NotificationRuleRow.created_at)
                .where(NotificationRuleRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return NotificationRuleRow.created_at > subquery

        return inner


class NotificationRuleOrders:
    """Query orders for notification rules."""

    @staticmethod
    def name(ascending: bool = True) -> QueryOrder:
        if ascending:
            return NotificationRuleRow.name.asc()
        else:
            return NotificationRuleRow.name.desc()

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return NotificationRuleRow.created_at.asc()
        else:
            return NotificationRuleRow.created_at.desc()

    @staticmethod
    def updated_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return NotificationRuleRow.updated_at.asc()
        else:
            return NotificationRuleRow.updated_at.desc()
